#include <Python.h>
#include "structmember.h"

#include "tiffio.h"
#include "tiffiop.h"
#include "tif_config.h"

#define MIN(X, Y) (((X) < (Y)) ? (X) : (Y))
#define MAX(X, Y) (((X) > (Y)) ? (X) : (Y))
#define TORGB(b, g, r) ((b << 16) | (g << 8) | r)
#define TORGBA(a, b, g, r) ((a << 24) | (b << 16) | (g << 8) | r)
#define GETR(p) ((p) & 0xff)
#define GETG(p) (((p) >> 8) & 0xff)
#define GETB(p) (((p) >> 16) & 0xff)
#define FIXUPPER(v) (((v) > (255)) ? (255) : (v))
#define FIXLOWER(v) (((v) < (0)) ? (0) : (v))
#define CLIP(x) ((x) <= 0 ? 0 : (x) < 256 ? (x) : 255)

typedef struct {
    PyObject_HEAD
    PyObject *src;
    PyObject *dest;
    PyObject *width;
    PyObject *height;
    PyObject *colors;
    PyObject *bits;
    PyObject *isTiff;
    PyObject *rhisto;
    PyObject *ghisto;
    PyObject *bhisto;
    PyObject *version_str;
} ImgProc;

static const char VERS[] = "0.1";
struct HSL rgbToHsl(int red, int green, int blue);
struct RGB hslToRgb(float hue, float sat, float lum);
static float HueToRGB(float v1, float v2, float vH);
struct Histo getHisto(uint32 * buffer, uint32 length);
static void setHistogram(ImgProc *self, struct Histo h);
static void getCumulativeHistogram(ImgProc *self, unsigned long * red, 
                                   unsigned long * green, unsigned long * blue);
static int getLowerThreshold(unsigned long * hist, int low);
static int getUpperThreshold(unsigned long * hist, int up);
static int transpose(ImgProc *self, const char *fname, double * mat);
static int getBuffer(const char *fname, uint32 * buffer, uint32 width, 
                     uint32 height);
static int saveBuffer(ImgProc *self, uint32 * buf);
static int saveRaster(ImgProc *self, uint32 ** raster);
static void reverseRaster(ImgProc *self, uint32 **raster);
static int flipArray(uint32 **arrayIn, uint32 **arrayOut, int width, int height);
static int mirrorArray(uint32 **arrayIn, uint32 **arrayOut, int width, int height);

/* Define a RGB pixel */
struct RGB
{
	unsigned char R;
	unsigned char G;
	unsigned char B;
};


/* Define a HSL pixel */
struct HSL
{
	int H;
	float S;
	float L;
};


/* Define the image histogram */
struct Histo
{
    unsigned int red[256];
    unsigned int green[256];
    unsigned int blue[256];
};


/* Conversion RGB to HSL */
struct HSL
rgbToHsl(int red, int green, int blue)
{
    struct HSL hsl;
    float r, g, b, min, max, delta, sat, hue, lum, threshold=0.5;

    r = red / 255;
    g = green / 255;
    b = blue / 255;
    min = MIN(r, MIN(g, b));
    max = MAX(r, MAX(g, b));
    lum = (float) (max + min) / 2.0;
    delta = (max - min);
    if (max == min) {
        hsl.H = 0;
        hsl.S = 0.0f;
        hsl.L = lum;
        return hsl;
    }
    sat = lum > threshold ? delta / (2 - max - min) : delta / (max + min);
    if (r == max) {
        hue = (float) (g - b) / delta;
        if (g < b) {
            hue = (int) hue % 6;
        }
    }
    else if (g == max) {
        hue = (b - r) / delta + 2; 
    }
    else {
        hue = (r - g) / delta + 4; 
    }

    hue = (int)(hue * 60.0);
    if (hue < 0) {
        hue += 360;
    }

    hsl.H = (int) hue;
    hsl.S = sat * 100;
    hsl.L = lum * 100;

    return hsl;
}


/* Conversion HSL to RGB */
struct RGB
hslToRgb(float hue, float sat, float lum)
{
    struct RGB rgb;
    if (sat == 0) {
        rgb.R = rgb.G = rgb.B = (unsigned char) lum;
        return rgb;
    }

	float v1, v2;
	hue /= 360;
	v2 = (lum < 0.5) ? (lum * (1 + sat)) : ((lum + sat) - (lum * sat));
	v1 = 2 * lum - v2;
	rgb.R = (unsigned char)(255 * HueToRGB(v1, v2, hue + (1.0f / 3)));
	rgb.G = (unsigned char)(255 * HueToRGB(v1, v2, hue));
	rgb.B = (unsigned char)(255 * HueToRGB(v1, v2, hue - (1.0f / 3)));

    return rgb;
}


/* Compute the R, G or b with the Hue */
static float 
HueToRGB(float v1, float v2, float hue)
{
	if (hue < 0) {
		hue += 1;
    }

	if (hue > 1) {
		hue -= 1;
    }

	if ((6 * hue) < 1) {
		return (v1 + (v2 - v1) * 6 * hue);
    }

	if ((2 * hue) < 1) {
		return v2;
    }

	if ((3 * hue) < 2) {
		return (v1 + (v2 - v1) * ((2.0f / 3) - hue) * 6);
    }

	return v1;
}

/* Return the RGB image histogram
 * buffer -- the RGBA buffer
 * length -- number of pixels
 */
struct Histo
getHisto(uint32 * buffer, uint32 length)
{
    struct Histo histo;
    uint r, g, b, i=0;

    for (i = 0; i < 256; i++) {
        histo.red[i] = 0;
        histo.green[i] = 0;
        histo.blue[i] = 0;
    }

    for (i = 0; i < length; i++) {
        r = (int )TIFFGetR(buffer[i]);
        g = (int )TIFFGetG(buffer[i]);
        b = (int )TIFFGetB(buffer[i]);
        histo.red[r]++;
        histo.green[g]++;
        histo.blue[b]++;
    }

    return histo;
}


/* Copy the image histogram into the struct ImgProc
 * h -- struct Histo
 */
static void
setHistogram(ImgProc *self, struct Histo h)
{
    int i;

    for (i = 0; i < 256; i++) {
        PyList_SET_ITEM(self->rhisto, i, PyLong_FromLong(h.red[i]));
        PyList_SET_ITEM(self->ghisto, i, PyLong_FromLong(h.green[i]));
        PyList_SET_ITEM(self->bhisto, i, PyLong_FromLong(h.blue[i]));
    }

}


/* Compute the cumulative histogram
 * red -- array for the red cumulative histogram
 * green -- array for the green cumulative histogram
 * blue -- array for the blue cumulative histogram
 */
static void
getCumulativeHistogram(ImgProc *self, unsigned long * red, unsigned long * green, 
                        unsigned long * blue)
{
    int i, j;
    red[0] = PyLong_AsUnsignedLong(PyList_GET_ITEM(self->rhisto, 0));
    green[0] = PyLong_AsUnsignedLong(PyList_GET_ITEM(self->ghisto, 0));
    blue[0] = PyLong_AsUnsignedLong(PyList_GET_ITEM(self->bhisto, 0));

    for (i = 0, j = 1; j < 256; i++, j++) {
        red[j] = red[i] + PyLong_AsUnsignedLong(
                    PyList_GET_ITEM(self->rhisto, j));
        green[j] = green[i] + PyLong_AsUnsignedLong(
                    PyList_GET_ITEM(self->ghisto, j));
        blue[j] = blue[i] + PyLong_AsUnsignedLong(
                    PyList_GET_ITEM(self->bhisto, j));
    }
}


/* Compute the lower threshold into the cumulative histogram
 * hist -- the cumulative histogram
 * low -- the threshold factor
 */
static int
getLowerThreshold(unsigned long * hist, int low)
{
    int i, min = 0;
    for (i = 0; i < 255; i++) {
        if ((int)hist[i] <= low) {
            min += 1;
        }
        else {
            break;
        }
    }
    return min;
}


/* Compute the upper threshold into the cumulative histogram
 * hist -- the cumulative histogram
 * low -- the threshold factor
 */
static int
getUpperThreshold(unsigned long * hist, int up)
{
    int max;
    for (max = 254; max > 0; max--) {
        if ((int)hist[max] > up) {
            break;
        }
    }
    max += 1;
    return max;
}


static int 
transpose(ImgProc *self, const char *fname, double * mat) 
{
    int red, green, blue, red1, green1, blue1, res=1;
    uint32 width = PyLong_AsUnsignedLong(self->width);
    uint32 height = PyLong_AsUnsignedLong(self->height);
    unsigned long i, npixels;
    uint32 *buffer;

    npixels = width * height;
    buffer = (uint32 *) _TIFFmalloc(npixels * sizeof(uint32));
    if (buffer == NULL) {
        const char *msg = "Can't create buffer";
        PyErr_SetString(PyExc_MemoryError, msg);
		return 0;
    }
         
    if (getBuffer(fname, buffer, width, height) == 0) {
        _TIFFfree(buffer);
        return 0;
    }

    for (i = 0; i < npixels; i++) {
        red = (int )TIFFGetR(buffer[i]);
        green = (int )TIFFGetG(buffer[i]);
        blue = (int )TIFFGetB(buffer[i]);

        red1 = (int ) (red * mat[0]) + (green * mat[1]) + (blue * mat[2]);
        green1 = (int ) (red * mat[3]) + (green * mat[4]) + (blue * mat[5]);
        blue1 = (int ) (red * mat[6]) + (green * mat[7]) + (blue * mat[8]);

        buffer[i] = TORGBA(255, FIXUPPER(blue1), 
                            FIXUPPER(green1), FIXUPPER(red1));
    }

    res = saveBuffer(self, buffer);
    return res;
}


/* Read the image as an RGBA 8 bits per sample
 * fname -- file name
 * buffer -- the buffer to write the data
 * width -- width of the image
 * height -- height of the image
 */
static int
getBuffer(const char *fname, uint32 * buffer, uint32 width, uint32 height)
{
	TIFF *tif = TIFFOpen(fname, "r");
	if (tif == NULL) {
        const char *msg = "Can't open the file";
        PyErr_SetString(PyExc_OSError, msg);
		return 0;
    }

    if (TIFFReadRGBAImage(tif, width, height, buffer, 0) != 1) {
        const char *msg = "Can't load the data";
        (void) TIFFClose(tif);
        PyErr_SetString(PyExc_MemoryError, msg);
		return 0;
    }

    (void) TIFFClose(tif);
    tif = NULL;
    return 1;
}


/* Write the image into a file
 * buf -- the data buffer
 */
static int
saveBuffer(ImgProc *self, uint32 * buf)
{
    int res = 1;
    uint32 width = PyLong_AsUnsignedLong(self->width);
    uint32 height = PyLong_AsUnsignedLong(self->height);
    uint16 v[1];
    unsigned long i, offset;
    const char *oname;
    tsize_t linebytes = 4 * width;
    uint32 *strip = PyMem_RawMalloc(width * sizeof(uint32));
    if (strip == NULL) {
        const char *msg = "saveBuffer can't create strip";
        PyErr_SetString(PyExc_MemoryError, msg);
		return 0;
    }

    offset = width * (height - 1);
    oname = PyUnicode_AsUTF8(self->dest);
    if (oname == NULL) {
        PyMem_RawFree(strip);
        const char *msg = "Can't read the out file name";
        PyErr_SetString(PyExc_OSError, msg);
        return 0;
    }

	TIFF *out = TIFFOpen(oname, "w");
	if (out == NULL) {
        PyMem_RawFree(strip);
        const char *msg = "Can't open the out file";
        PyErr_SetString(PyExc_OSError, msg);
		return 0;
    }

    v[0] = EXTRASAMPLE_ASSOCALPHA;
    TIFFSetField(out, TIFFTAG_IMAGEWIDTH, width);
    TIFFSetField(out, TIFFTAG_IMAGELENGTH, height);
    TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 4);
    TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField(out, TIFFTAG_EXTRASAMPLES, 1, v);
    TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
    TIFFSetField(out, TIFFTAG_ROWSPERSTRIP, TIFFDefaultStripSize(out, linebytes));

    for (i = 0; i < height; i++) {
        memcpy(strip, buf + offset, linebytes);    
        if (TIFFWriteScanline(out, strip, i, 0) < 0) {
            res = 0;
            break;
        }
        offset -= width;
    }

    PyMem_RawFree(strip);
    (void) TIFFClose(out);
    out = NULL;

    return res;
}


/* Write the image into a file
 * buf -- the data raster
 */
static int
saveRaster(ImgProc *self, uint32 ** raster)
{
    int res = 1;
    uint32 width = PyLong_AsUnsignedLong(self->width);
    uint32 height = PyLong_AsUnsignedLong(self->height);
    uint16 v[1];
    unsigned long i;
    const char *oname;
    tsize_t linebytes = 4 * width;
    uint32 *strip = PyMem_RawMalloc(width * sizeof(uint32));
    if (strip == NULL) {
        const char *msg = "saveRaster can't create strip";
        PyErr_SetString(PyExc_MemoryError, msg);
		return 0;
    }

    oname = PyUnicode_AsUTF8(self->dest);
    if (oname == NULL) {
        PyMem_RawFree(strip);
        const char *msg = "Can't read the out file name";
        PyErr_SetString(PyExc_OSError, msg);
        return 0;
    }

	TIFF *out = TIFFOpen(oname, "w");
	if (out == NULL) {
        PyMem_RawFree(strip);
        const char *msg = "Can't open the out file";
        PyErr_SetString(PyExc_OSError, msg);
		return 0;
    }

    v[0] = EXTRASAMPLE_ASSOCALPHA;
    TIFFSetField(out, TIFFTAG_IMAGEWIDTH, width);
    TIFFSetField(out, TIFFTAG_IMAGELENGTH, height);
    TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 4);
    TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, 8);
    TIFFSetField(out, TIFFTAG_EXTRASAMPLES, 1, v);
    TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
    TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
    TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
    TIFFSetField(out, TIFFTAG_ROWSPERSTRIP, TIFFDefaultStripSize(out, linebytes));

    for (i = 0; i < height; i++) {
        memcpy(strip, (uint32*) raster[i], linebytes);    
        if (TIFFWriteScanline(out, strip, i, 0) < 0) {
            res = 0;
            break;
        }
    }

    PyMem_RawFree(strip);
    (void) TIFFClose(out);
    out = NULL;

    return res;
}

/* Reverse the color values
 * 0 --> 255
 * 1 --> 254
 * 2 --> 253 ......
 */
static void
reverseRaster(ImgProc *self, uint32 **raster)
{
    uint32 width = PyLong_AsUnsignedLong(self->width);
    uint32 height = PyLong_AsUnsignedLong(self->height);
    uint32 pixel;
    int red, green, blue; 
    uint row, col;
    for (row = 0; row < height; row++) {
        for (col = 0; col < width; col++) {
            pixel = raster[row][col];
            red = TIFFGetR(pixel);
            green = TIFFGetG(pixel);
            blue = TIFFGetB(pixel);
            raster[row][col] = TORGBA(255, (255-blue), (255-green), (255-red));
        }
    }
}


/* Flip an array 2D
 * arrayIn -- source array
 * arrayOut -- array flipped
 * width -- width of array
 * height -- height of array
 */
static int 
flipArray(uint32 **arrayIn, uint32 **arrayOut, int width, int height)
{
    int row, out = height-1;
    int length = width * 4;
    for (row = 0; row < height; row++) {
        _TIFFmemcpy(arrayOut[out], arrayIn[row], length);
        out--;
    }

    int red = TIFFGetR(arrayIn[0][0]);
    int green = TIFFGetG(arrayIn[0][0]);
    int blue = TIFFGetB(arrayIn[0][0]);
    printf("In: %d, %d, %d\n", red, green, blue);
    red = TIFFGetR(arrayOut[height-1][0]);
    green = TIFFGetG(arrayOut[height-1][0]);
    blue = TIFFGetB(arrayOut[height-1][0]);
    printf("out: %d, %d, %d\n", red, green, blue);
    return 1;
}

/* Mirrorize an array 2D
 * arrayIn -- source array
 * arrayOut -- array mirrorized
 * width -- width of array
 * height -- height of array
 */
static int 
mirrorArray(uint32 **arrayIn, uint32 **arrayOut, int width, int height)
{
    int row, col, out = width-1;
    printf("Mirroring: %d, %d\n", width, height);
    for (row = 0; row < height; row++) {
        for (col = 0; col < width; col++) {
            arrayOut[row][out] = (uint32) arrayIn[row][col];
            out--;
        }
        out = width-1;
    }

    return 1;
}

static PyObject *
ImgProc_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    ImgProc *self;
    Py_ssize_t length = 256;

    self = (ImgProc *)type->tp_alloc(type, 0);
    if (self != NULL){
        self->src = PyUnicode_FromString("");
        if (self->src == NULL) {
            Py_DECREF(self);
            return NULL;
        }

        self->dest = PyUnicode_FromString("");
        if (self->dest == NULL) {
            Py_DECREF(self);
            return NULL;
        }

        self->version_str = PyUnicode_FromString(VERS);
        self->width = PyLong_FromLong(0);
        self->height = PyLong_FromLong(0);
        self->colors = PyLong_FromLong(0);
        self->bits = PyLong_FromLong(0);
        self->isTiff = PyLong_FromLong(0);
        self->rhisto = PyList_New(length);
        self->ghisto = PyList_New(length);
        self->bhisto = PyList_New(length);
    }

    return (PyObject *)self;
}


static PyObject *
ImgProc_init(ImgProc *self, PyObject *args, PyObject *kwds)
{
    PyObject *tmp;
    char *fname;

    if (!PyArg_ParseTuple(args, "s", &fname))
        return NULL;

    if (strlen(fname) > 4) {
        tmp = self->src;
        self->src = PyUnicode_FromString(fname);
        Py_XDECREF(tmp);
    }
    return 0;
}


static void
ImgProc_dealloc(ImgProc* self)
{
    Py_XDECREF(self->src);
    Py_XDECREF(self->dest);
    Py_XDECREF(self->width);
    Py_XDECREF(self->height);
    Py_XDECREF(self->colors);
    Py_XDECREF(self->bits);
    Py_XDECREF(self->isTiff);
    Py_XDECREF(self->rhisto);
    Py_XDECREF(self->ghisto);
    Py_XDECREF(self->bhisto);
    Py_XDECREF(self->version_str);
}


static PyObject *
ImgProc_readImage(ImgProc *self)
{
    const char *iname;
    uint32 width;
    uint32 height;
    short bits;
    short samples;

    iname = PyUnicode_AsUTF8(self->src);
    if (iname == NULL) {
        const char *msg = "Can't read the file name";
        PyErr_SetString(PyExc_OSError, msg);
        return NULL;
    }

	TIFF *tif = TIFFOpen(iname, "r");
	if (tif == NULL) {
        const char *msg = "Can't open the file";
        PyErr_SetString(PyExc_OSError, msg);
		return NULL;
    }

    if (TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width) < 1) {
        const char *msg = "Can't read the width of the image";
        PyErr_SetString(PyExc_OSError, msg);
        (void) TIFFClose(tif);
		return NULL;
    }

    if (TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height) < 1) {
        const char *msg = "Can't read the height of the image";
        PyErr_SetString(PyExc_OSError, msg);
        (void) TIFFClose(tif);
		return NULL;
    }

    if (TIFFGetField(tif, TIFFTAG_BITSPERSAMPLE, &bits) < 1) {
        const char *msg = "Can't read the bits per sample of the image";
        PyErr_SetString(PyExc_OSError, msg);
        (void) TIFFClose(tif);
		return NULL;
    }
    if (TIFFGetField(tif, TIFFTAG_SAMPLESPERPIXEL, &samples) < 1) {
        const char *msg = "Can't read the colors number of the image";
        PyErr_SetString(PyExc_OSError, msg);
        (void) TIFFClose(tif);
		return NULL;
    }

    (void) TIFFClose(tif);
    tif = NULL;

    self->width = PyLong_FromLong(width);
    if (self->width == NULL) {
        const char *msg = "Unvalid width value";
        PyErr_SetString(PyExc_ValueError, msg);
        return NULL;
    }

    self->height = PyLong_FromLong(height);
    if (self->height == NULL) {
        const char *msg = "Unvalid height value";
        PyErr_SetString(PyExc_ValueError, msg);
        return NULL;
    }

    self->bits = PyLong_FromLong(bits);
    if (self->bits == NULL) {
        const char *msg = "Unvalid bits number value";
        PyErr_SetString(PyExc_ValueError, msg);
        return NULL;
    }

    self->colors = PyLong_FromLong(samples);
    if (self->colors == NULL) {
        const char *msg = "Unvalid colors number value";
        PyErr_SetString(PyExc_ValueError, msg);
        return NULL;
    }

    self->isTiff = PyLong_FromLong(1);
    Py_RETURN_NONE;
}

static PyObject *
ImgProc_getHistogram(ImgProc *self)
{
    struct Histo histo;
    uint32 width = PyLong_AsUnsignedLong(self->width);
    uint32 height = PyLong_AsUnsignedLong(self->height);
    unsigned long npixels = width * height;
    uint32 *buffer;
    const char *iname;

    iname = PyUnicode_AsUTF8(self->src);
    if (iname == NULL) {
        const char *msg = "Can't read the file name";
        PyErr_SetString(PyExc_OSError, msg);
        return NULL;
    }

    npixels = width * height;
    buffer = (uint32 *) _TIFFmalloc(npixels * sizeof(uint32));
    if (buffer == NULL) {
        const char *msg = "Can't create buffer";
        PyErr_SetString(PyExc_MemoryError, msg);
		return NULL;
    }

    if (getBuffer(iname, buffer, width, height) == 0) {
        _TIFFfree(buffer);
        return NULL;
    }

    histo = getHisto(buffer, npixels);
    setHistogram(self, histo);
    _TIFFfree(buffer);
    Py_RETURN_NONE;
}


static PyObject *
ImgProc_contrastBrightnessFilter(ImgProc *self, PyObject *args)
{
    float gain, bias, lower, upper, tmp;
    if (!PyArg_ParseTuple(args, "ffff", &gain, &bias, &lower, &upper)) {
        return NULL;
    }

    if (PyLong_AsUnsignedLong(self->isTiff) == 0) {
        const char *msg = "Not a valid TIFF image";
        PyErr_SetString(PyExc_TypeError, msg);
        return NULL;
    }

    int red, green, blue;
    uint32 width = PyLong_AsUnsignedLong(self->width);
    uint32 height = PyLong_AsUnsignedLong(self->height);
    unsigned long i, npixels;
    const char *iname;
    uint32 *buffer;

    iname = PyUnicode_AsUTF8(self->src);
    if (iname == NULL) {
        const char *msg = "Can't read the file name";
        PyErr_SetString(PyExc_OSError, msg);
        return NULL;
    }

    npixels = width * height;
    buffer = (uint32 *) _TIFFmalloc(npixels * sizeof(uint32));
    if (buffer == NULL) {
        const char *msg = "Can't create buffer";
        PyErr_SetString(PyExc_MemoryError, msg);
		return NULL;
    }
         
    if (getBuffer(iname, buffer, width, height) == 0) {
        _TIFFfree(buffer);
        return NULL;
    }

    for (i = 0; i < npixels; i++) {
        red = (int )TIFFGetR(buffer[i]);
        green = (int )TIFFGetG(buffer[i]);
        blue = (int )TIFFGetB(buffer[i]);

        tmp = (float) red;
        if (tmp > lower && tmp < upper) {
            tmp = tmp * gain + bias;
            red = (int) tmp;
            red = CLIP(red);
        }

        tmp = (float) green;
        if (tmp > lower && tmp < upper) {
            tmp = tmp * gain + bias;
            green = (int) tmp;
            green = CLIP(green);
        }

        tmp = (float) blue;
        if (tmp > lower && tmp < upper) {
            tmp = tmp * gain + bias;
            blue = (int) tmp;
            blue = CLIP(blue);
        }

        buffer[i] = TORGBA(255, blue, green, red);
        
    }
    if (saveBuffer(self, buffer) == 0) {
        const char *msg = "Image writing failed";
        _TIFFfree(buffer);
        PyErr_SetString(PyExc_OSError, msg);
		return NULL;
    }

    _TIFFfree(buffer);
    Py_RETURN_NONE;
}


static PyObject *
ImgProc_contrastStretchingFilter(ImgProc *self, PyObject *args)
{
    int lower, upper;
    if (!PyArg_ParseTuple(args, "ii", &lower, &upper)) {
        return NULL;
    }

    if (PyLong_AsUnsignedLong(self->isTiff) == 0) {
        const char *msg = "Not a valid TIFF image";
        PyErr_SetString(PyExc_TypeError, msg);
        return NULL;
    }

    printf("contrastStretchingFilter lower: %d, upper: %d\n", lower, upper);
    uint8 r, g, b;
    unsigned int rmin, rmax, gmin, gmax, bmin, bmax;
    uint32 width = PyLong_AsUnsignedLong(self->width);
    uint32 height = PyLong_AsUnsignedLong(self->height);
    unsigned long i, npixels;
    const char *iname;
    uint32 *buffer;
    unsigned long cred[256] = {0};
    unsigned long cgreen[256] = {0};
    unsigned long cblue[256] = {0};

    iname = PyUnicode_AsUTF8(self->src);
    if (iname == NULL) {
        const char *msg = "Can't read the file name";
        PyErr_SetString(PyExc_OSError, msg);
        return NULL;
    }

    npixels = width * height;
    buffer = (uint32 *) _TIFFmalloc(npixels * sizeof(uint32));
    if (buffer == NULL) {
        const char *msg = "Can't create buffer";
        PyErr_SetString(PyExc_MemoryError, msg);
		return NULL;
    }

    if (getBuffer(iname, buffer, width, height) == 0) {
        _TIFFfree(buffer);
        return NULL;
    }

    getCumulativeHistogram(self, cred, cgreen, cblue);
    rmin = getLowerThreshold(cred, lower);
    gmin = getLowerThreshold(cgreen, lower);
    bmin = getLowerThreshold(cblue, lower);
    rmax = getUpperThreshold(cred, upper);
    gmax = getUpperThreshold(cgreen, upper);
    bmax = getUpperThreshold(cblue, upper);

    int rnorm[256] = {0};
    if (rmax > rmin) {
        for (i = rmin; i < rmax; i++) {
            rnorm[i] = ((i - rmin) * 255) / ((rmax - rmin) + 0.5);
        }

        for(i = rmax; i < 256; i++)
            rnorm[i] = 255;
    }
    else {
        for (i = 0; i < 256; i++) {
            rnorm[i] = 127;
        }
    }

    int gnorm[256] = {0};
    if (gmax > gmin) {
        for (i = gmin; i < gmax; i++) {
            gnorm[i] = ((i - gmin) * 255) / ((gmax - gmin) + 0.5);
        }

        for(i = gmax; i < 256; i++) {
            gnorm[i] = 255;
        }
    }
    else {
        for (i = 0; i < 256; i++) {
            gnorm[i] = 127;
        }
    }

    int bnorm[256] = {0};
    if (bmax > bmin) {
        for (i = bmin; i < bmax; i++) {
            bnorm[i] = ((i - bmin) * 255) / ((bmax - bmin) + 0.5);
        }

        for(i = bmax; i < 256; i++) {
            bnorm[i] = 255;
        }
    }
    else {
        for (i = 0; i < 256; i++) {
            bnorm[i] = 127;
        }
    }

    for (i = 0; i < npixels; i++) {
        r = (int )TIFFGetR(buffer[i]);
        g = (int )TIFFGetG(buffer[i]);
        b = (int )TIFFGetB(buffer[i]);
        buffer[i] = TORGBA(255, bnorm[b], gnorm[g], rnorm[r]);
    }

    if (saveBuffer(self, buffer) == 0) {
        const char *msg = "Image writing failed";
        _TIFFfree(buffer);
        PyErr_SetString(PyExc_OSError, msg);
		return NULL;
    }

    _TIFFfree(buffer);
    Py_RETURN_NONE;
}


static PyObject *
ImgProc_sharpen(ImgProc *self, PyObject *args)
{
    int sumr, sumg, sumb, x, y, t, err, center;
    uint32 *buffer;
    const char *iname;
    uint32 width = PyLong_AsUnsignedLong(self->width);
    uint32 height = PyLong_AsUnsignedLong(self->height);
    unsigned long row, col, npixels, offset;

    if (!PyArg_ParseTuple(args, "i", &center)) {
        return NULL;
    }

    int kernel[9] = {0, -1, 0, -1, center, -1, 0, -1, 0};
    iname = PyUnicode_AsUTF8(self->src);
    if (iname == NULL) {
        const char *msg = "Can't read the file name";
        PyErr_SetString(PyExc_OSError, msg);
        return NULL;
    }

    npixels = width * height;
    offset = npixels - width;
    uint32  length = 4 * width;
    buffer = _TIFFmalloc(npixels * sizeof(uint32));
    if (buffer == NULL) {
        const char *msg = "Can't create buffer";
        PyErr_SetString(PyExc_MemoryError, msg);
		return NULL;
    }

    if (getBuffer(iname, buffer, width, height) == 0) {
        _TIFFfree(buffer);
        return NULL;
    }

    uint32 *raster[height];
    for (row = 0; row < height; row++) {
        raster[row] = _TIFFmalloc(width * sizeof(uint32));
    }

    for (row = 0; row < height; row++) {
        _TIFFmemcpy(raster[row], buffer + offset, length);
        offset -= width;
    }

    uint32 *rasterOut[height];
    for (row = 0; row < height; row++) {
        rasterOut[row] = _TIFFmalloc(width * sizeof(uint32));
    }

    offset = npixels - width;
    for (row = 0; row < height; row++) {
        _TIFFmemcpy(raster[row], buffer + offset, length);
        offset -= width;
    }
    _TIFFfree(buffer);

    for (row = 1; row < height - 1; row++) {
        for (col = 1; col < width - 1; col++) {
            sumr = 0;
            sumg = 0;
            sumb = 0;
            t = 0;
            for (y = -1; y < 2; y++) {
                for (x = -1; x < 2; x++) {
                    sumr += (int) (kernel[t] * TIFFGetR(raster[row+y][col+x]));
                    sumg += (int) (kernel[t] * TIFFGetG(raster[row+y][col+x]));
                    sumb += (int) (kernel[t] * TIFFGetB(raster[row+y][col+x]));
                    t++;
                }
            }
            sumr = CLIP(sumr);
            sumg = CLIP(sumg);
            sumb = CLIP(sumb);
            rasterOut[row][col] = TORGBA(255, sumb, sumg, sumr);
        }
    }

    err = saveRaster(self, rasterOut);
    for (row = 0; row < height; row++) {
        _TIFFfree(raster[row]);
        raster[row] = NULL;
    }

    for (row = 0; row < height; row++) {
        //printf("%lu ", row);
        _TIFFfree(rasterOut[row]);
        rasterOut[row] = NULL;
    }

    if (err == 0) {
        const char *msg = "Image writing failed";
        PyErr_SetString(PyExc_OSError, msg);
		return NULL;
    }

    Py_RETURN_NONE;
}
            

static PyObject *
ImgProc_laplacianFilter(ImgProc *self, PyObject *args)
{
    char a, b, c, d, e, f, g, h, i;
    int x, y, sumr, sumg, sumb, err, reverse=0, t=0;
    const char *iname;
    unsigned long row, col, npixels, offset;
    uint32 *buffer;
    float mean, gap;

    if (!PyArg_ParseTuple(args, "iiiiiiiiiff|i", &a, &b, &c, &d, &e, &f, &g, &h, 
                          &i, &mean, &gap, &reverse)) {
        return NULL;
    }

    /* Avoid division by zero */
    if (mean == 0) {
        mean = 1;
    }

    int template[9] = {a, b, c, d, e, f, g, h, i};
    uint32 width = PyLong_AsUnsignedLong(self->width);
    uint32 height = PyLong_AsUnsignedLong(self->height);
    iname = PyUnicode_AsUTF8(self->src);
    if (iname == NULL) {
        const char *msg = "Can't read the file name";
        PyErr_SetString(PyExc_OSError, msg);
        return NULL;
    }

    npixels = width * height;
    offset = npixels - width;
    uint32  length = 4 * width;
    buffer = _TIFFmalloc(npixels * sizeof(uint32));
    if (buffer == NULL) {
        const char *msg = "Can't create buffer";
        PyErr_SetString(PyExc_MemoryError, msg);
		return NULL;
    }

    if (getBuffer(iname, buffer, width, height) == 0) {
        _TIFFfree(buffer);
        return NULL;
    }

    uint32 *raster[height];
    for (row = 0; row < height; row++) {
        raster[row] = _TIFFmalloc(width * sizeof(uint32));
    }

    for (row = 0; row < height; row++) {
        _TIFFmemcpy(raster[row], buffer + offset, length);
        offset -= width;
    }
    _TIFFfree(buffer);

    uint32 *rasterOut[height];
    for (row = 0; row < height; row++) {
        rasterOut[row] = _TIFFmalloc(width * sizeof(uint32));
    }

    for (row = 1; row < height - 1; row++) {
        for (col = 1; col < width - 1; col++) {
            sumr = 0;
            sumg = 0;
            sumb = 0;
            t = 8;
            for (y = -1; y < 2; y++) {
                for (x = -1; x < 2; x++) {
                    sumr += (int) (template[t] * TIFFGetR(raster[row+y][col+x]));
                    sumg += (int) (template[t] * TIFFGetG(raster[row+y][col+x]));
                    sumb += (int) (template[t] * TIFFGetB(raster[row+y][col+x]));
                    t--;
                }
            }

            sumr = (int) (sumr / mean + gap);
            sumg = (int) (sumg / mean + gap);
            sumb = (int) (sumb / mean + gap);
            sumr = CLIP(sumr);
            sumg = CLIP(sumg);
            sumb = CLIP(sumb);

            rasterOut[row][col] = TORGBA(255, sumb, sumg, sumr);
        }
        /* Duplicate first and last pixels */
        rasterOut[row][0] = rasterOut[row][1];
        rasterOut[row][col] = rasterOut[row][col-1];
    }
    /* Duplicate first and last rows */
    _TIFFmemcpy(rasterOut[0], rasterOut[1], length);
    _TIFFmemcpy(rasterOut[height-1], rasterOut[height-2], length);

    if (reverse){
        reverseRaster(self, rasterOut);
    }

    err = saveRaster(self, rasterOut);
    for (row = 0; row < height; row++) {
        _TIFFfree(raster[row]);
        raster[row] = NULL;
    }

    for (row = 0; row < height; row++) {
        //printf("%lu ", row);
        _TIFFfree(rasterOut[row]);
        rasterOut[row] = NULL;
    }

    if (err == 0) {
        const char *msg = "Image writing failed";
        PyErr_SetString(PyExc_OSError, msg);
		return NULL;
    }

    Py_RETURN_NONE;

}


static PyObject *
ImgProc_greyMapFilter(ImgProc *self, PyObject *args)
{
    int level;
    const char *iname;

    if (!PyArg_ParseTuple(args, "i", &level)) {
        return NULL;
    }

    if (PyLong_AsUnsignedLong(self->isTiff) == 0) {
        const char *msg = "Not a valid TIFF image";
        PyErr_SetString(PyExc_TypeError, msg);
        return NULL;
    }

    printf("Apply grey map filter level %d\n", level);
    iname = PyUnicode_AsUTF8(self->src);
    if (iname == NULL) {
        const char *msg = "Can't read the file name";
        PyErr_SetString(PyExc_OSError, msg);
        return NULL;
    }

    if (level == 0) {
        double mat[9] = {0.34, 0.7, 0.02, 0.34, 0.7, 0.02, 0.34, 0.7, 0.02};
        transpose(self, iname, mat);
    }
    else if (level == 1) {
        double mat[9] = {0.02, 0.7, 0.34, 0.02, 0.7, 0.34, 0.02, 0.7, 0.34};
        transpose(self, iname, mat);
    }
    else if (level == 2) {
        double mat[9] = {0.14, 0.67, 0.38, 0.14, 0.67, 0.38, 0.14, 0.67, 0.38};
        transpose(self, iname, mat);
    }
    else if (level == 3) {
        double mat[9] = {0.32, 0.57, 0.38, 0.32, 0.57, 0.38, 0.32, 0.57, 0.38};
        transpose(self, iname, mat);
    }
    else if (level == 4) {
        double mat[9] = {0.05, 0.6, 0.6, 0.05, 0.6, 0.6, 0.05, 0.6, 0.6};
        transpose(self, iname, mat);
    }
    else {
        double mat[9] = {0.03, 0.76, 0.54, 0.03, 0.76, 0.54, 0.03, 0.76, 0.54};
        transpose(self, iname, mat);
    }

    Py_RETURN_NONE;
}


static PyObject *
ImgProc_sepiaToningFilter(ImgProc *self, PyObject *args)
{
    int level;
    const char *iname;

    if (!PyArg_ParseTuple(args, "i", &level)) {
        return NULL;
    }

    if (PyLong_AsUnsignedLong(self->isTiff) == 0) {
        const char *msg = "Not a valid TIFF image";
        PyErr_SetString(PyExc_TypeError, msg);
        return NULL;
    }

    printf("Apply grey map filter level %d\n", level);
    iname = PyUnicode_AsUTF8(self->src);
    if (iname == NULL) {
        const char *msg = "Can't read the file name";
        PyErr_SetString(PyExc_OSError, msg);
        return NULL;
    }

    if (level == 0) {
        double mat[9] = {0.8, 0.69, 0.56, 0.46, 0.37, 0.22, 0.45, 0.4, 0.25};
        transpose(self, iname, mat);
    }
    else if (level == 1) {
        double mat[9] = {0.85, 0.78, 0.6, 0.51, 0.42, 0.32, 0.43, 0.41, 0.31};
        transpose(self, iname, mat);
    }
    else if (level == 2) {
        double mat[9] = {0.79, 0.72, 0.62, 0.6, 0.51, 0.4, 0.48, 0.42, 0.36};
        transpose(self, iname, mat);
    }
    else if (level == 3) {
        double mat[9] = {0.83, 0.72, 0.62, 0.63, 0.54, 0.43, 0.48, 0.41, 0.36};
        transpose(self, iname, mat);
    }
    else {
        double mat[9] = {0.86, 0.75, 0.65, 0.65, 0.56, 0.45, 0.51, 0.45, 0.39};
        transpose(self, iname, mat);
    }

    Py_RETURN_NONE;
}


static PyObject *
ImgProc_blueToningFilter(ImgProc *self, PyObject *args)
{
    int level;
    const char *iname;

    if (!PyArg_ParseTuple(args, "i", &level)) {
        return NULL;
    }

    if (PyLong_AsUnsignedLong(self->isTiff) == 0) {
        const char *msg = "Not a valid TIFF image";
        PyErr_SetString(PyExc_TypeError, msg);
        return NULL;
    }

    printf("Apply grey map filter level %d\n", level);
    iname = PyUnicode_AsUTF8(self->src);
    if (iname == NULL) {
        const char *msg = "Can't read the file name";
        PyErr_SetString(PyExc_OSError, msg);
        return NULL;
    }

    if (level == 0) {
        double mat[9] = {0.27, 0.53, 0.13, 0.35, 0.59, 0.17, 0.4, 0.77, 0.3};
        transpose(self, iname, mat);
    }
    else if (level == 1) {
        double mat[9] = {0.29, 0.55, 0.15, 0.4, 0.73, 0.22, 0.5, 0.77, 0.29};
        transpose(self, iname, mat);
    }
    else if (level == 2) {
        double mat[9] = {0.32, 0.58, 0.18, 0.4, 0.73, 0.22, 0.45, 0.82, 0.24};
        transpose(self, iname, mat);
    }
    else if (level == 3) {
        double mat[9] = {0.34, 0.6, 0.2, 0.45, 0.78, 0.27, 0.5, 0.77, 0.29};
        transpose(self, iname, mat);
    }
    else {
        double mat[9] = {0.38, 0.64, 0.24, 0.49, 0.82, 0.31, 0.54, 0.81, 0.33};
        transpose(self, iname, mat);
    }

    Py_RETURN_NONE;
}


static PyObject *
ImgProc_colorTemperature(ImgProc *self, PyObject *args)
{
    int red, green, blue, rd, gr, bl, min, max, sr, sg, sb;
    float strength, strg, ol;
    uint32 *buffer;
    unsigned long i, npixels;
    const char *iname;

    if (!PyArg_ParseTuple(args, "iiif", &red, &green, &blue, &strength)) {
        return NULL;
    }

    strg = 1 - strength;
    uint32 width = PyLong_AsUnsignedLong(self->width);
    uint32 height = PyLong_AsUnsignedLong(self->height);
    iname = PyUnicode_AsUTF8(self->src);
    if (iname == NULL) {
        const char *msg = "Can't read the file name";
        PyErr_SetString(PyExc_OSError, msg);
        return NULL;
    }

    npixels = width * height;
    buffer = _TIFFmalloc(npixels * sizeof(uint32));
    if (buffer == NULL) {
        const char *msg = "Can't create buffer";
        PyErr_SetString(PyExc_MemoryError, msg);
		return NULL;
    }

    if (getBuffer(iname, buffer, width, height) == 0) {
        _TIFFfree(buffer);
        return NULL;
    }

    Py_BEGIN_ALLOW_THREADS
    for (i = 0; i < npixels; i++) {
        rd = (int )TIFFGetR(buffer[i]);
        gr = (int )TIFFGetG(buffer[i]);
        bl = (int )TIFFGetB(buffer[i]);

        /* Keep the original luminance */
        min = fmin(bl, fmin(gr, rd));
        max = fmax(bl, fmax(gr, rd));
        ol = (min + max) / 2;

        sr = (rd * strg) + (red * strength);
        sg = (gr * strg) + (green * strength);
        sb = (bl * strg) + (blue * strength);

        struct HSL hsl = rgbToHsl(sr, sg, sb);
        /* Reset the original luminance */
        hsl.L = ol;
        struct RGB rgb = hslToRgb(hsl.H, hsl.S, hsl.L);

        buffer[i] = TORGBA(255, CLIP(rgb.B), CLIP(rgb.G), CLIP(rgb.R));
    }
    Py_END_ALLOW_THREADS

    saveBuffer(self, buffer);
    _TIFFfree(buffer);
    Py_RETURN_NONE;
}


static PyObject *
ImgProc_flipImage(ImgProc *self)
{
    uint32 width = PyLong_AsUnsignedLong(self->width);
    uint32 height = PyLong_AsUnsignedLong(self->height);
    uint row;
    const char *iname;
    iname = PyUnicode_AsUTF8(self->src);
    if (iname == NULL) {
        const char *msg = "Can't read the file name";
        PyErr_SetString(PyExc_OSError, msg);
        return NULL;
    }

    uint32 *buffer;
    uint32 npixels = width * height;
    uint32 offset = npixels - width;
    uint32 length = 4 * width;

    buffer = _TIFFmalloc(npixels * sizeof(uint32));
    if (buffer == NULL) {
        const char *msg = "Can't create buffer";
        PyErr_SetString(PyExc_MemoryError, msg);
		return NULL;
    }

    if (getBuffer(iname, buffer, width, height) == 0) {
        _TIFFfree(buffer);
        return NULL;
    }

    uint32 *raster[height];
    for (row = 0; row < height; row++) {
        raster[row] = _TIFFmalloc(width * sizeof(uint32));
    }

    for (row = 0; row < height; row++) {
        _TIFFmemcpy(raster[row], buffer + offset, length);
        offset -= width;
    }
    _TIFFfree(buffer);

    uint32 *rasterOut[height];
    for (row = 0; row < height; row++) {
        rasterOut[row] = _TIFFmalloc(width * sizeof(uint32));
    }

    int res = flipArray(raster, rasterOut, width, height);
    if (res == 0) {
        const char *msg = "Flip array failed";
        PyErr_SetString(PyExc_RuntimeError, msg);
		return NULL;
    }

    int err = saveRaster(self, rasterOut);

    for (row = 0; row < height; row++) {
        _TIFFfree(raster[row]);
        raster[row] = NULL;
    }

    for (row = 0; row < height; row++) {
        _TIFFfree(rasterOut[row]);
        rasterOut[row] = NULL;
    }

    if (err == 0) {
        const char *msg = "Image writing failed";
        PyErr_SetString(PyExc_OSError, msg);
		return NULL;
    }

    Py_RETURN_NONE;

}


static PyObject *
ImgProc_mirrorizeImage(ImgProc *self)
{
    uint32 width = PyLong_AsUnsignedLong(self->width);
    uint32 height = PyLong_AsUnsignedLong(self->height);
    uint row;
    const char *iname;
    iname = PyUnicode_AsUTF8(self->src);
    if (iname == NULL) {
        const char *msg = "Can't read the file name";
        PyErr_SetString(PyExc_OSError, msg);
        return NULL;
    }

    uint32 *buffer;
    uint32 npixels = width * height;
    uint32 offset = npixels - width;
    uint32 length = 4 * width;
    buffer = _TIFFmalloc(npixels * sizeof(uint32));
    if (buffer == NULL) {
        const char *msg = "Can't create buffer";
        PyErr_SetString(PyExc_MemoryError, msg);
		return NULL;
    }

    if (getBuffer(iname, buffer, width, height) == 0) {
        _TIFFfree(buffer);
        return NULL;
    }

    uint32 *raster[height];
    for (row = 0; row < height; row++) {
        raster[row] = _TIFFmalloc(width * sizeof(uint32));
    }

    for (row = 0; row < height; row++) {
        _TIFFmemcpy(raster[row], buffer + offset, length);
        offset -= width;
    }
    _TIFFfree(buffer);

    uint32 *rasterOut[height];
    for (row = 0; row < height; row++) {
        rasterOut[row] = _TIFFmalloc(width * sizeof(uint32));
    }

    int res = mirrorArray(raster, rasterOut, width, height);
    if (res == 0) {
        const char *msg = "Flip array failed";
        PyErr_SetString(PyExc_RuntimeError, msg);
		return NULL;
    }

    int err = saveRaster(self, rasterOut);

    for (row = 0; row < height; row++) {
        _TIFFfree(raster[row]);
        raster[row] = NULL;
    }

    for (row = 0; row < height; row++) {
        _TIFFfree(rasterOut[row]);
        rasterOut[row] = NULL;
    }

    if (err == 0) {
        const char *msg = "Image writing failed";
        PyErr_SetString(PyExc_OSError, msg);
		return NULL;
    }

    Py_RETURN_NONE;

}


static PyMemberDef ImgProc_members[] = {
    {"source_path", T_OBJECT_EX, offsetof(ImgProc, src), 0,
     "Path of the tiff source file"},
    {"dest_path", T_OBJECT_EX, offsetof(ImgProc, dest), 0,
     "Path of the tiff destination file"},
    {"width", T_OBJECT_EX, offsetof(ImgProc, width), 0,
     "The image width, before any rotation"},
    {"height", T_OBJECT_EX, offsetof(ImgProc, height), 0,
     "The image height, before any rotation"},
    {"colors", T_OBJECT_EX, offsetof(ImgProc, colors), 0,
     "The number of samples per pixel"},
    {"bits", T_OBJECT_EX, offsetof(ImgProc, bits), 0,
     "The number of bits par sample"},
    {"version", T_OBJECT_EX, offsetof(ImgProc, version_str), 0,
     "The ImgProc version number as string"},
    {"histo_red", T_OBJECT_EX, offsetof(ImgProc, rhisto), 0,
     "The values of the red histogram"},
    {"histo_green", T_OBJECT_EX, offsetof(ImgProc, ghisto), 0,
     "The values of the green histogram"},
    {"histo_blue", T_OBJECT_EX, offsetof(ImgProc, bhisto), 0,
     "The values of the blue histogram"},
    {"is_tiff", T_OBJECT_EX, offsetof(ImgProc, isTiff), 0,
     "Return True if the file is a valid tiff"},
    {NULL}  /* Sentinel */
};

static PyMethodDef ImgProc_methods[] = {
    {"read", (PyCFunction)ImgProc_readImage, METH_NOARGS,
     "Read the image"},
    {"getHistogram", (PyCFunction)ImgProc_getHistogram, METH_VARARGS,
     "Compute the color histogram of the image"},
    {"applyContrastBrightness", (PyCFunction)ImgProc_contrastBrightnessFilter, 
     METH_VARARGS, "Enhance the contrast-brightness of the image"},
    {"applyContrastStretching", (PyCFunction)ImgProc_contrastStretchingFilter,
     METH_VARARGS, "Enhance the contrast of the image"},
    {"sharpen", (PyCFunction)ImgProc_sharpen, METH_VARARGS,
     "Enforce the sharpen of the image"},
    {"applyLaplacian", (PyCFunction)ImgProc_laplacianFilter, METH_VARARGS,
     "Apply a Laplacian filter to the image"},
    {"applyGreyMap", (PyCFunction)ImgProc_greyMapFilter, METH_VARARGS,
     "Apply a grey map filter to the image"},
    {"applySepiaToning", (PyCFunction)ImgProc_sepiaToningFilter, METH_VARARGS,
     "Apply a sepia toning filter to the image"},
    {"applyBlueToning", (PyCFunction)ImgProc_blueToningFilter, METH_VARARGS,
     "Apply a blue toning filter to the image"},
    {"applyColorTemp", (PyCFunction)ImgProc_colorTemperature, 
     METH_VARARGS, "Change the color temperature of the image"},
    {"flipImage", (PyCFunction)ImgProc_flipImage, METH_NOARGS,
     "Flip the image"},
    {"mirrorizeImage", (PyCFunction)ImgProc_mirrorizeImage, METH_NOARGS,
     "Flip the image"},
    {NULL, NULL, 0, NULL}        /* Sentinel */
};

static PyTypeObject ImgProcType = {
    PyVarObject_HEAD_INIT(NULL, 0)
    "_liboqapy.ImgProc",          /* tp_name */
    sizeof(ImgProc),              /* tp_basicsize */
    0,                            /* tp_itemsize */
    (destructor)ImgProc_dealloc,  /* tp_dealloc */
    0,                            /* tp_print */
    0,                            /* tp_getattr */
    0,                            /* tp_setattr */
    0,                            /* tp_reserved */
    0,                            /* tp_repr */
    0,                            /* tp_as_number */
    0,                            /* tp_as_sequence */
    0,                            /* tp_as_mapping */
    0,                            /* tp_hash  */
    0,                            /* tp_call */
    0,                            /* tp_str */
    0,                            /* tp_getattro */
    0,                            /* tp_setattro */
    0,                            /* tp_as_buffer */
    Py_TPFLAGS_DEFAULT |
        Py_TPFLAGS_BASETYPE,      /* tp_flags */
    "ImgProc object",             /* tp_doc */
    0,                            /* tp_traverse */
    0,                            /* tp_clear */
    0,                            /* tp_richcompare */
    0,                            /* tp_weaklistoffset */
    0,                            /* tp_iter */
    0,                            /* tp_iternext */
    ImgProc_methods,              /* tp_methods */
    ImgProc_members,              /* tp_members */
    0,                            /* tp_getset */
    0,                            /* tp_base */
    0,                            /* tp_dict */
    0,                            /* tp_descr_get */
    0,                            /* tp_descr_set */
    0,                            /* tp_dictoffset */
    (initproc)ImgProc_init,       /* tp_init */
    0,                            /* tp_alloc */
    ImgProc_new,                  /* tp_new */
};

static struct PyModuleDef liboqapymodule = {
    PyModuleDef_HEAD_INIT,
    "_liboqapy",                 /* name of module */
    "Image processing module",   /* module documentation, may be NULL */
    -1,                          /* size of per-interpreter state of the module,
                     or -1 if the module keeps state in global variables. */
    ImgProc_methods
};

PyMODINIT_FUNC
PyInit__liboqapy(void)
{
    PyObject* m;

    if (PyType_Ready(&ImgProcType) < 0)
        return NULL;

    m = PyModule_Create(&liboqapymodule);
    if (m == NULL)
        return NULL;

    Py_INCREF(&ImgProcType);
    PyModule_AddObject(m, "ImgProc", (PyObject *)&ImgProcType);
    return m;
}

