# -*- coding: utf-8 -*-

# layouttoolbox.py
# this file is part of Oqapy serie 3
#
# Author: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
# Home: http://www.oqapy.eu
# Copyright: 2010-2019 Vincent Vande Vyvre
# Licence: GPL3
#
# layout tool box


import logging
lgg = logging.getLogger(__name__)

from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QIcon, QPixmap
from PyQt5.QtWidgets import (QWidget, QGridLayout, QVBoxLayout, QHBoxLayout, 
                             QComboBox, QToolButton, QPushButton, QSpinBox,
                             QSizePolicy, QSpacerItem, QLabel, QFrame,
                             QDoubleSpinBox, QCheckBox, QRadioButton)

import stylesheets as styles


class LayoutToolBox:
    def __init__(self, ui):
        self.ui = ui
        self.widget = ToolBox(self)
        self.on_reset = False
        self.size = None
        self.old_size = None
        self.old_preset = 0
        self.item_resized = False
        self.margins = [0, 0, 0, 0]
        self.widget.mm_rdo.toggled.connect(self.on_unity_changed)
        self.widget.preset_size_cmb.currentIndexChanged.connect(
                                self.on_preset_size_changed)
        self.widget.width_spb.valueChanged.connect(self.on_width_changed)
        self.widget.height_spb.valueChanged.connect(self.on_height_changed)
        self.widget.undo_btn.clicked.connect(self.on_undo_clicked)
        self.widget.apply_resize_btn.clicked.connect(self.apply_resize)
        self.widget.preset_pos_cmb.currentIndexChanged.connect(
                                self.on_preset_pos_changed)
        self.widget.top_spb.valueChanged.connect(self.on_top_changed)
        self.widget.left_spb.valueChanged.connect(self.on_left_changed)
        self.widget.bottom_spb.valueChanged.connect(self.on_bottom_changed)
        self.widget.right_spb.valueChanged.connect(self.on_right_changed)
        self.widget.up_btn.clicked.connect(self.on_up_clicked)
        self.widget.down_btn.clicked.connect(self.on_down_clicked)

    @property
    def factor(self):
        if self.size is None:
            return 1

        return self.size[1] / self.size[0]

    @property
    def is_proportionnal(self):
        return self.widget.proportion_chb.isChecked()

    @property
    def unity(self):
        return self.widget.mm_rdo.isChecked()

    def reset(self):
        w = self.widget
        self.on_reset = True
        w.preset_size_cmb.setCurrentIndex(0)
        w.width_spb.setValue(10)
        w.height_spb.setValue(10)
        w.undo_btn.setEnabled(False)
        w.apply_resize_btn.setEnabled(False)
        w.preset_pos_cmb.setCurrentIndex(0)
        w.top_spb.setValue(0.0)
        w.left_spb.setValue(0.0)
        w.bottom_spb.setValue(0.0)
        w.right_spb.setValue(0.0)
        self.size = None
        self.on_reset = False
        self.old_preset = 0
        self.item_resized = False

    def active_widget(self, active):
        if not active:
            self.reset()
        self.widget.setEnabled(active)

    def update_with_item(self, item, factor, preset=False):
        self.current_item = item
        self.ratio = factor
        w = self.widget
        self.block_all_signals(True)
        pos = item.pos()
        p = item.pixmap()
        s = self.ui.scene.sceneRect()
        width = int(p.width()/factor)
        height = int(p.height()/factor)
        top = pos.y() / factor
        left = pos.x() / factor
        bottom = (s.height() - (pos.y() + p.height())) / factor
        right = (s.width() - (pos.x() + p.width())) / factor
        if not self.unity:
            width /= 25.4
            height /= 25.4
            top /= 25.4
            left /= 25.4
            bottom /= 25.4
            right /= 25.4

        w.width_spb.setValue(width)
        w.height_spb.setValue(height)
        w.apply_resize_btn.setEnabled(False)
        w.top_spb.setValue(top)
        w.left_spb.setValue(left)
        w.bottom_spb.setValue(bottom)
        w.right_spb.setValue(right)
        self.size = (width, height)
        if not self.item_resized:
            w.apply_resize_btn.setEnabled(False)
            self.old_size = self.size

        if preset:
            w.preset_pos_cmb.setCurrentIndex(0)

        self.block_all_signals(False)
        self.configure_order_buttons(item.level)

    def update_size_and_pos(self):
        w = self.widget
        p = self.current_item.pixmap()
        self.block_all_signals(True)
        suffix = (" in.", " mm.")[self.unity]
        if self.unity:
            w.width_spb.setValue(int(p.width()/self.ratio))
            w.height_spb.setValue(int(p.height()/self.ratio))

        else:
            new = self.convert_unity(p.width()/self.ratio)
            w.width_spb.setValue(new)
            new = self.convert_unity(p.height()/self.ratio)
            w.height_spb.setValue(new)

        w.width_spb.setSuffix(suffix)
        w.height_spb.setSuffix(suffix)
        for item in (w.top_spb, w.left_spb, w.bottom_spb, w.right_spb):
            new = self.convert_unity(item.value())
            item.setValue(new)
            item.setSuffix(suffix)

        self.block_all_signals(False)

    def configure_order_buttons(self, z, max_=None):
        self.widget.down_btn.setEnabled(z>0)
        if max_ is None:
            max_ = len(self.ui.main.render.items) - 1

        self.widget.up_btn.setEnabled(z<max_)

    def block_size_signals(self, bool_):
        self.widget.width_spb.blockSignals(bool_)
        self.widget.height_spb.blockSignals(bool_)

    def block_margins_signals(self, bool_):
        self.widget.top_spb.blockSignals(bool_)
        self.widget.left_spb.blockSignals(bool_)
        self.widget.bottom_spb.blockSignals(bool_)
        self.widget.right_spb.blockSignals(bool_)

    def block_all_signals(self, bool_):
        self.block_size_signals(bool_)
        self.block_margins_signals(bool_)
        self.widget.preset_pos_cmb.blockSignals(bool_)

    def set_minimal_margins(self, margins):
        self.margins = margins

    def on_preset_size_changed(self, idx):
        if not self.on_reset:
            self.old_size = (self.widget.width_spb.value(), 
                             self.widget.height_spb.value())
            self.item_resized = True
            self.spy = 2
            self.ui.fit_item(idx)
            self.set_preset_size_index(idx)
            self.widget.undo_btn.setEnabled(True)

    def set_preset_size_index(self, index):
        self.widget.preset_size_cmb.blockSignals(True)
        self.widget.preset_size_cmb.setCurrentIndex(index)
        self.widget.preset_size_cmb.blockSignals(False)

    def on_unity_changed(self, state):
        self.update_size_and_pos()

    def on_width_changed(self, value):
        if not self.on_reset:
            if self.widget.proportion_chb.isChecked():
                self.widget.height_spb.blockSignals(True)
                self.widget.height_spb.setValue(value*self.factor)
                self.widget.height_spb.blockSignals(False)

            self.widget.undo_btn.setEnabled(True)
            self.widget.apply_resize_btn.setEnabled(True)

    def on_height_changed(self, value):
        if not self.on_reset:
            if self.widget.proportion_chb.isChecked():
                self.widget.width_spb.blockSignals(True)
                self.widget.width_spb.setValue(value/self.factor)
                self.widget.width_spb.blockSignals(False)

            self.widget.undo_btn.setEnabled(True)
            self.widget.apply_resize_btn.setEnabled(True)

    def on_undo_clicked(self):
        self.block_size_signals(True)
        self.widget.width_spb.setValue(self.old_size[0])
        self.widget.height_spb.setValue(self.old_size[1])
        self.block_size_signals(False)
        if self.item_resized:
            self.apply_resize(0)
        self.widget.undo_btn.setEnabled(False)
        self.widget.apply_resize_btn.setEnabled(False)

    def apply_resize(self, idx):
        self.widget.apply_resize_btn.setEnabled(False)
        width = self.widget.width_spb.value()
        height = self.widget.height_spb.value()
        if not self.unity:
            width *= 25.4
            height *= 25.4

        if self.is_proportionnal:
            self.ui.scene.resize_item(int(width))

        else:
            self.ui.scene.resize_item(int(width), int(height))

    def convert_unity(self, value):
        if self.unity:
            # inch to mm
            return int(value * 25.4)

        return value / 25.4

    def on_preset_pos_changed(self, idx):
        if not self.on_reset:
            self.ui.scene.move_item(idx)

    def on_top_changed(self, value):
        if not self.on_reset:
            self.ui.scene.set_item_top_position(value)

    def on_right_changed(self, value):
        if not self.on_reset:
            self.ui.scene.set_item_right_position(value)

    def on_bottom_changed(self, value):
        if not self.on_reset:
            self.ui.scene.set_item_bottom_position(value)

    def on_left_changed(self, value):
        if not self.on_reset:
            self.ui.scene.set_item_left_position(value)

    def on_up_clicked(self):
        self.ui.main.render.set_item_z_order(1)

    def on_down_clicked(self):
        self.ui.main.render.set_item_z_order(-1)


class ToolBox(QWidget):
    def __init__(self, driver):
        super().__init__(None)
        self.driver = driver
        self.gridLayout_2 = QGridLayout(self)
        self.gridLayout_2.setContentsMargins(2, 0, 10, 0)
        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setSpacing(0)
        self.gridLayout = QGridLayout()
        self.gridLayout.setVerticalSpacing(2)
        self.hlayout = QHBoxLayout()
        spacerItem7 = QSpacerItem(20, 20, QSizePolicy.Fixed, 
                                  QSizePolicy.Minimum)
        self.hlayout.addItem(spacerItem7)
        self.mm_rdo = QRadioButton(self)
        self.mm_rdo.setText(_("Millimeter"))
        self.mm_rdo.setChecked(True)
        self.hlayout.addWidget(self.mm_rdo)
        self.inch_rdo = QRadioButton(self)
        self.inch_rdo.setText(_("Inch"))
        self.hlayout.addWidget(self.inch_rdo)
        self.gridLayout.addLayout(self.hlayout, 0, 0, 1, 2)
        self.label_1 = QLabel(self)
        self.label_1.setAlignment(Qt.AlignRight|Qt.AlignTrailing|
                                  Qt.AlignVCenter)
        self.gridLayout.addWidget(self.label_1, 1, 0, 1, 2)
        self.preset_size_cmb = QComboBox(self)
        self.gridLayout.addWidget(self.preset_size_cmb, 1, 2, 1, 1)
        self.label_2 = QLabel(self)
        self.label_2.setAlignment(Qt.AlignRight|Qt.AlignTrailing|
                                  Qt.AlignVCenter)
        self.gridLayout.addWidget(self.label_2, 2, 0, 1, 2)
        self.width_spb = QDoubleSpinBox(self)
        #self.width_spb.setMinimum(10)
        self.width_spb.setMaximum(10000)
        self.gridLayout.addWidget(self.width_spb, 2, 2, 1, 1)
        self.label_3 = QLabel(self)
        self.label_3.setAlignment(Qt.AlignRight|Qt.AlignTrailing|
                                  Qt.AlignVCenter)
        self.gridLayout.addWidget(self.label_3, 3, 0, 1, 2)
        self.height_spb = QDoubleSpinBox(self)
        #self.height_spb.setMinimum(10)
        self.height_spb.setMaximum(10000)
        self.gridLayout.addWidget(self.height_spb, 3, 2, 1, 1)
        self.proportion_chb = QCheckBox(self)
        self.proportion_chb.setChecked(True)
        self.gridLayout.addWidget(self.proportion_chb, 4, 1, 1, 1)
        self.hlayout2 = QHBoxLayout()
        spacerItem8 = QSpacerItem(20, 20, QSizePolicy.Expanding, 
                                  QSizePolicy.Minimum)
        self.hlayout2.addItem(spacerItem8)
        self.undo_btn = QPushButton(self)
        icon = QIcon()
        icon.addPixmap(QPixmap("medias/edit-undo.png"), 
                       QIcon.Normal, QIcon.Off)
        self.undo_btn.setIcon(icon)
        self.undo_btn.setIconSize(QSize(20, 20))
        self.undo_btn.setEnabled(False)
        self.hlayout2.addWidget(self.undo_btn)
        self.apply_resize_btn = QPushButton(self)
        icon1 = QIcon()
        icon1.addPixmap(QPixmap("medias/dialog-apply.svg"), 
                       QIcon.Normal, QIcon.Off)
        self.apply_resize_btn.setIcon(icon1)
        self.apply_resize_btn.setIconSize(QSize(20, 20))
        self.apply_resize_btn.setEnabled(False)
        self.hlayout2.addWidget(self.apply_resize_btn)
        self.gridLayout.addLayout(self.hlayout2, 5, 0, 1, 3)
        self.line = QFrame(self)
        self.line.setFrameShape(QFrame.HLine)
        self.line.setFrameShadow(QFrame.Sunken)
        self.gridLayout.addWidget(self.line, 6, 0, 1, 3)
        self.label_4 = QLabel(self)
        self.label_4.setAlignment(Qt.AlignRight|Qt.AlignTrailing|
                                  Qt.AlignVCenter)
        self.gridLayout.addWidget(self.label_4, 7, 1, 1, 1)
        self.preset_pos_cmb = QComboBox(self)
        self.gridLayout.addWidget(self.preset_pos_cmb, 7, 2, 1, 2)
        self.label_5 = QLabel(self)
        self.label_5.setAlignment(Qt.AlignRight|Qt.AlignTrailing|
                                  Qt.AlignVCenter)
        self.gridLayout.addWidget(self.label_5, 8, 1, 1, 1)
        spacerItem = QSpacerItem(40, 20, QSizePolicy.Expanding, 
                                 QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 9, 2, 1, 1)
        self.verticalLayout_2.addLayout(self.gridLayout)
        self.horizontalLayout_4 = QHBoxLayout()
        spacerItem1 = QSpacerItem(28, 20, QSizePolicy.Expanding, 
                                  QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem1)
        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setSpacing(2)
        self.horizontalLayout = QHBoxLayout()
        spacerItem2 = QSpacerItem(40, 20, QSizePolicy.Expanding, 
                                  QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem2)
        self.top_spb = QDoubleSpinBox(self)
        self.top_spb.setRange(-1000, 9000)
        self.horizontalLayout.addWidget(self.top_spb)
        spacerItem3 = QSpacerItem(40, 20, QSizePolicy.Expanding, 
                                  QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem3)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.horizontalLayout_2 = QHBoxLayout()
        self.left_spb = QDoubleSpinBox(self)
        self.left_spb.setRange(-1000, 9000)
        self.horizontalLayout_2.addWidget(self.left_spb)
        self.right_spb = QDoubleSpinBox(self)
        self.right_spb.setRange(-1000, 9000)
        self.horizontalLayout_2.addWidget(self.right_spb)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3 = QHBoxLayout()
        spacerItem4 = QSpacerItem(40, 20, QSizePolicy.Expanding, 
                                  QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem4)
        self.bottom_spb = QDoubleSpinBox(self)
        self.bottom_spb.setRange(-1000, 9000)
        self.horizontalLayout_3.addWidget(self.bottom_spb)
        spacerItem5 = QSpacerItem(40, 20, QSizePolicy.Expanding, 
                                  QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem5)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_4.addLayout(self.verticalLayout)
        self.verticalLayout_2.addLayout(self.horizontalLayout_4)
        self.horizontalLayout_5 = QHBoxLayout()
        spacerItem6 = QSpacerItem(40, 20, QSizePolicy.Fixed, 
                                  QSizePolicy.Minimum)
        self.horizontalLayout_5.addItem(spacerItem6)
        self.zorder_lbl = QLabel(self)
        self.horizontalLayout_5.addWidget(self.zorder_lbl)
        self.up_btn = QPushButton(self)
        icon2 = QIcon()
        icon2.addPixmap(QPixmap("medias/go-up.png"), QIcon.Normal, QIcon.Off)
        self.up_btn.setIcon(icon2)
        self.horizontalLayout_5.addWidget(self.up_btn)
        self.down_btn = QPushButton(self)
        icon3 = QIcon()
        icon3.addPixmap(QPixmap("medias/go-down.png"), QIcon.Normal, QIcon.Off)
        self.down_btn.setIcon(icon3)
        self.horizontalLayout_5.addWidget(self.down_btn)
        self.verticalLayout_2.addLayout(self.horizontalLayout_5)
        self.gridLayout_2.addLayout(self.verticalLayout_2, 0, 0, 1, 1)
        self.up_btn.setEnabled(False)
        self.down_btn.setEnabled(False)

        self.retranslateUi()
        self.set_style()

        self.units = ["mm", "Point", "in", "Pica", "Didot", "Cicero", 
                        "Device pixel"]
        # Populate comboBoxs
        ico_path = ["medias/center.png", "medias/full_page.png", 
                    "medias/full_width.png", "medias/full_height.png"]
        texts = ["Undefined", "Full page", "Full width", "Full height"]

        for i, p in enumerate(ico_path):
            icon = QIcon()
            icon.addPixmap(QPixmap(p), QIcon.Normal, QIcon.Off)
            self.preset_size_cmb.addItem(icon, texts[i])

        ico_path = ["medias/top_left.png", "medias/top_right.png", 
                    "medias/center.png", "medias/bottom_left.png", 
                    "medias/bottom_right.png"]
        texts = ["Top left", "Top right", "Center", "Bottom left", 
                    "Bottom right"]
        self.preset_pos_cmb.addItem("Undefined")

        for i, p in enumerate(ico_path):
            icon = QIcon()
            icon.addPixmap(QPixmap(p), QIcon.Normal, QIcon.Off)
            self.preset_pos_cmb.addItem(icon, texts[i])

    def retranslateUi(self):
        self.label_1.setText(_("Adjustment: "))
        self.label_2.setText(_("width: "))
        self.width_spb.setSuffix(_(" mm."))
        self.label_3.setText(_("Height: "))
        self.height_spb.setSuffix(_(" mm."))
        self.proportion_chb.setText(_("Proportionnal"))
        self.undo_btn.setToolTip(_("Reset the old size"))
        self.undo_btn.setText(_("Undo"))
        self.apply_resize_btn.setToolTip(_("Apply the new size"))
        self.apply_resize_btn.setText(_("Apply"))
        self.label_4.setText(_("Position: "))
        self.label_5.setText(_("Margins:"))
        self.top_spb.setSuffix(_(" mm."))
        self.top_spb.setToolTip(_("Top margin"))
        self.left_spb.setSuffix(_(" mm."))
        self.left_spb.setToolTip(_("Left margin"))
        self.right_spb.setSuffix(_(" mm."))
        self.right_spb.setToolTip(_("Right margin"))
        self.bottom_spb.setSuffix(_(" mm."))
        self.bottom_spb.setToolTip(_("Bottom margin"))
        self.zorder_lbl.setText(_("Z Order: "))
        self.up_btn.setText(_("Up"))
        self.down_btn.setText(_("Down"))

    def set_style(self):
        self.mm_rdo.setStyleSheet(styles.RADIO)
        self.inch_rdo.setStyleSheet(styles.RADIO)
        self.preset_size_cmb.setStyleSheet(styles.COMBO)
        self.width_spb.setStyleSheet(styles.DSPINBOX)
        self.height_spb.setStyleSheet(styles.DSPINBOX)
        self.proportion_chb.setStyleSheet(styles.CHECKBOX)
        self.undo_btn.setStyleSheet(styles.PUSHBUTTON)
        self.apply_resize_btn.setStyleSheet(styles.PUSHBUTTON)
        self.preset_pos_cmb.setStyleSheet(styles.COMBO)
        self.top_spb.setStyleSheet(styles.DSPINBOX)
        self.left_spb.setStyleSheet(styles.DSPINBOX)
        self.right_spb.setStyleSheet(styles.DSPINBOX)
        self.bottom_spb.setStyleSheet(styles.DSPINBOX)
        self.up_btn.setStyleSheet(styles.PUSHBUTTON)
        self.down_btn.setStyleSheet(styles.PUSHBUTTON)


