# -*- coding: utf-8 -*-

# previewtoolbox.py
# this file is part of Oqapy serie 3
#
# Author: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
# Home: http://www.oqapy.eu
# Copyright: 2010-2019 Vincent Vande Vyvre
# Licence: GPL3
#
# image preview tool box


import os
import logging
lgg = logging.getLogger(__name__)

from PyQt5.QtCore import Qt, QSize, QRectF
from PyQt5.QtGui import QIcon, QPixmap, QImage, QTransform
from PyQt5.QtWidgets import (QWidget, QGridLayout, QVBoxLayout, QHBoxLayout,
                             QToolButton, QPushButton, QGraphicsView, 
                             QGraphicsScene, QSpacerItem, QSizePolicy)

import stylesheets as styles


class PreviewerToolBox:
    def __init__(self, ui, parent):
        self.ui = ui
        self.widget = ToolBox(self, parent)
        self.widget.left_btn.clicked.connect(self.rotate_left)
        self.widget.right_btn.clicked.connect(self.rotate_right)
        self.widget.previous_btn.clicked.connect(self.previous_image)
        self.widget.next_btn.clicked.connect(self.next_image)
        self.widget.add_btn.clicked.connect(self.add_image)
        self.qimage = False

    def get_viewer_size(self):
        """Returns the size of graphics view."""
        return (self.widget.preview.width(), self.widget.preview.height())

    def configure(self, imgs, idx):
        self.images = imgs
        self.current = idx - 1
        self.count = len(self.images)
        self.scene = QGraphicsScene()
        self.widget.preview.setScene(self.scene)
        self.next_image()

    def previous_image(self):
        self.current = (self.current - 1) % self.count
        self.set_image(self.images[self.current])

    def next_image(self):
        self.current = (self.current + 1) % self.count
        self.set_image(self.images[self.current])

    def add_image(self):
        self.ui.main.add_new_image_item(self.qimage, self.images[self.current])

    def set_image(self, path):
        self.qimage = QImage(path)
        self.set_pixmap()
        self.set_title()

    def set_pixmap(self):
        w, h = self.get_viewer_size()
        pix = QPixmap(self.qimage.scaled(w-2, h-2, 
                            Qt.KeepAspectRatio, 
                            Qt.SmoothTransformation))
        self.change_preview(pix)

    def set_title(self, title=None):
        if title is None:
            title = os.path.basename(self.images[self.current])

        self.widget.parent().setTitle(title)

    def change_preview(self, pixmap):
        self.scene.clear()
        self.scene.setSceneRect(QRectF(pixmap.rect()))
        self.scene.addPixmap(pixmap)

    def rotate_left(self):
        self.rotate(-90)

    def rotate_right(self):
        self.rotate(90)

    def rotate(self, angle):
        matrix = QTransform()
        self.qimage = self.qimage.transformed(matrix.rotate(angle), 
                             Qt.SmoothTransformation)

        self.set_pixmap()

    def on_resize_event(self):
        if self.qimage:
            # qimage does not exists at startup
            self.set_pixmap()

    @property
    def current_image(self):
        return self.images[self.current]


class ToolBox(QWidget):
    def __init__(self, driver, parent=None):
        super().__init__(parent)
        self.driver = driver
        self.gridLayout_2 = QGridLayout(self)
        self.gridLayout_2.setContentsMargins(10, 0, 10, 0)
        self.gridLayout_2.setHorizontalSpacing(4)
        self.gridLayout_2.setVerticalSpacing(2)
        self.horizontalLayout = QHBoxLayout()
        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setContentsMargins(2, 6, 2, 6)
        self.verticalLayout.setSpacing(4)
        self.gridLayout = QGridLayout()
        self.gridLayout.setSpacing(2)
        self.left_btn = QToolButton(self)
        self.left_btn.setToolTip(_("Rotate to left"))
        icon = QIcon()
        icon.addPixmap(QPixmap("medias/left_arrow_20x20.png"), 
                        QIcon.Normal, QIcon.Off)
        self.left_btn.setIcon(icon)
        self.gridLayout.addWidget(self.left_btn, 0, 0, 1, 1)
        self.right_btn = QToolButton(self)
        self.right_btn.setToolTip(_("Rotate to right"))
        icon1 = QIcon()
        icon1.addPixmap(QPixmap("medias/right_arrow_20x20.png"), 
                        QIcon.Normal, QIcon.Off)
        self.right_btn.setIcon(icon1)
        self.gridLayout.addWidget(self.right_btn, 0, 1, 1, 1)
        self.previous_btn = QToolButton(self)
        self.previous_btn.setToolTip(_("Previous image"))
        icon2 = QIcon()
        icon2.addPixmap(QPixmap("medias/collapse_white_right.png"), 
                        QIcon.Normal, QIcon.Off)
        self.previous_btn.setIcon(icon2)
        self.gridLayout.addWidget(self.previous_btn, 1, 0, 1, 1)
        self.next_btn = QToolButton(self)
        self.next_btn.setToolTip(_("Next image"))
        icon3 = QIcon()
        icon3.addPixmap(QPixmap("medias/collapse_white_left.png"), 
                        QIcon.Normal, QIcon.Off)
        self.next_btn.setIcon(icon3)
        self.gridLayout.addWidget(self.next_btn, 1, 1, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.add_btn = QPushButton(self)
        self.add_btn.setMaximumSize(QSize(60, 16777215))
        self.add_btn.setToolTip(_("Add this image into the composition"))
        icon4 = QIcon()
        icon4.addPixmap(QPixmap("medias/add.png"), QIcon.Normal, QIcon.Off)
        self.add_btn.setIcon(icon4)
        self.verticalLayout.addWidget(self.add_btn)
        sp = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)
        self.verticalLayout.addItem(sp)
        self.horizontalLayout.addLayout(self.verticalLayout)
        self.preview = QGraphicsView(self)
        self.horizontalLayout.addWidget(self.preview)
        self.gridLayout_2.addLayout(self.horizontalLayout, 0, 0, 1, 1)
        self.set_style()

    def resizeEvent(self, event):
        self.driver.on_resize_event()

    def set_style(self):
        self.preview.setStyleSheet(styles.VIEWER)
        self.left_btn.setStyleSheet(styles.TBUTTON)
        self.right_btn.setStyleSheet(styles.TBUTTON)
        self.previous_btn.setStyleSheet(styles.TBUTTON)
        self.next_btn.setStyleSheet(styles.TBUTTON)
        self.add_btn.setStyleSheet(styles.PUSHBUTTON)

