# -*- coding: utf-8 -*-

# printertoolbox.py
# this file is part of Oqapy serie 3
#
# Author: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
# Home: http://www.oqapy.eu
# Copyright: 2010-2019 Vincent Vande Vyvre
# Licence: GPL3
#
# printer tool box


import os
import logging
lgg = logging.getLogger(__name__)

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QPixmap, QPageSize
from PyQt5.QtWidgets import (QWidget, QGridLayout, QVBoxLayout, QHBoxLayout, 
                             QComboBox, QToolButton, QPushButton, QSpinBox,
                             QLineEdit, QSizePolicy, QSpacerItem, QLabel,
                             QFileDialog, QMessageBox)
from PyQt5.QtPrintSupport import QPrinter, QPageSetupDialog

import stylesheets as styles


class PrinterToolBox:
    def __init__(self, ui):
        self.ui = ui
        self.widget = ToolBox(self)
        self.widget.output_cmb.currentTextChanged.connect(self.on_device_changed)
        self.widget.settings_btn.clicked.connect(self.change_paper_size)
        self.widget.apply_reso_btn.setEnabled(False)
        self.widget.papersize_btn.clicked.connect(self.change_paper_size)
        self.widget.orient_cmb.currentIndexChanged.connect(
                                            self.on_orientation_changed)
        self.widget.reso_spb.valueChanged.connect(self.on_resolution_changed)
        self.widget.apply_reso_btn.clicked.connect(self.on_apply_resolution)
        self.widget.browse_btn.clicked.connect(self.get_file_name)
        self.widget.print_btn.clicked.connect(self.on_print_clicked)

    def set_printer_parameters(self, printer):
        """Fix the parameters of the current printer.

        Args:
        printer_info -- instance of QPrinterInfo or None
        """
        if printer is None:
            self.set_pdf_parameters()

        else:
            self.resolution = printer.resolution()
            self.widget.output_cmb.clear()
            self.widget.output_cmb.addItems(self.ui.main.printers)
            self.widget.output_cmb.addItem("To file pdf")
            self.widget.output_cmb.setCurrentText(printer.name)
            self.set_page_size(printer)
            self.set_orientation(printer)
            self.widget.reso_spb.setValue(self.resolution)

    def set_page_size(self, device):
        size = device.paper_size
        txt = "%s x %s" %(size[0], size[1])
        self.widget.papersize_btn.setText(txt)

    def reverse_size(self):
        w, h = self.widget.papersize_btn.text().split(" x ")
        txt = "%s x %s" %(h, w)
        self.widget.papersize_btn.setText(txt)

    def set_orientation(self, device):
        self.widget.orient_cmb.blockSignals(True)
        self.widget.orient_cmb.setCurrentIndex(device.orientation_)
        self.widget.orient_cmb.blockSignals(False)

    def change_paper_size(self):
        ps = QPageSetupDialog(self.ui.main.current_printer)
        ret = ps.exec_()
        self.ui.main.printer_parameters_changed()

    def on_device_changed(self, device):
        if "pdf" in device:
            bool_ = True

        else:
            bool_ = False

        self.widget.fname_led.setEnabled(bool_)
        self.widget.browse_btn.setEnabled(bool_)

    def get_file_name(self):
        f = self.ui.main.current_folder
        name = QFileDialog.getSaveFileName(self.ui, "Save File", f,
                                           "Files (*.pdf)")[0]

        if name:
            path, ext = os.path.splitext(name)
            if not ext:
                name = path + self.get_output_format()

            self.widget.fname_led.setText(name)
            return name

        return False

    def get_output_format(self):
        txt = self.widget.output_cmb.currentText()
        if "pdf" in txt:
            return ".pdf"

    def ask_replace(self, f):
        """Show message if the file already exists in destination's folder.

        Args:
        f -- file name

        Return:
        True if file must be rewrite, otherwise False
        """
        mssg = QMessageBox()
        mssg.setIcon(QMessageBox.Question)
        mssg.setText(_("A file with the same name already exists in the" 
                    " destination folder."))
        mssg.setInformativeText(_("Do you want to replace the existing file?"))
        can = QPushButton(_("Cancel"))
        mssg.addButton(can, QMessageBox.ActionRole)
        rmp = QPushButton(_("Replace"))
        mssg.addButton(rmp, QMessageBox.ActionRole)
        mssg.setDefaultButton(can)
        reply = mssg.exec_()

        if reply == 0:
            return False

        return True

    def on_orientation_changed(self, index):
        self.ui.main._on_orientation_changed(index)
        self.reverse_size()

    def on_resolution_changed(self, value):
        if value != self.resolution:
            self.widget.apply_reso_btn.setEnabled(True)

        else:
            self.widget.apply_reso_btn.setEnabled(False)

    def on_apply_resolution(self):
        self.resolution = self.widget.reso_spb.value()
        self.widget.apply_reso_btn.setEnabled(False)
        self.ui.main.change_resolution(self.resolution)

    def on_print_clicked(self):
        fname = ""
        color = self.widget.color_cmb.currentIndex()
        copies = self.widget.copies_spb.value()
        if self.get_output_format() == ".pdf":
            fname = self.check_filename()
            if not fname:
                return

        self.ui.main.send_to_printer(color, copies, fname)

    def check_filename(self):
        name = self.widget.fname_led.text()
        if not name:
            name = self.get_file_name()

        if name:
            if os.path.isfile(name):
                if not self.ask_replace(name):
                    return False

        return name


class ToolBox(QWidget):
    def __init__(self, driver):
        super().__init__(None)
        self.driver = driver
        self.grid = QGridLayout(self)
        self.grid.setContentsMargins(20, 0, 10, 0)
        self.grid.setVerticalSpacing(2)
        self.grid.setHorizontalSpacing(6)
        self.verticalLayout = QVBoxLayout()
        self.gridLayout = QGridLayout()
        self.gridLayout.setVerticalSpacing(2)
        self.output_lbl = QLabel(self)
        self.output_lbl.setText(_("Output: "))
        self.output_lbl.setAlignment(Qt.AlignRight|Qt.AlignTrailing|
                                     Qt.AlignVCenter)
        self.gridLayout.addWidget(self.output_lbl, 0, 0, 1, 1)
        self.horizontalLayout = QHBoxLayout()
        self.output_cmb = QComboBox(self)
        self.output_cmb.setSizeAdjustPolicy(QComboBox.AdjustToContents)
        self.horizontalLayout.addWidget(self.output_cmb)
        self.settings_btn = QToolButton(self)
        self.settings_btn.setToolTip("Settings")
        icon = QIcon()
        icon.addPixmap(QPixmap("medias/settings.png"), QIcon.Normal, 
                       QIcon.Off)
        self.settings_btn.setIcon(icon)
        self.horizontalLayout.addWidget(self.settings_btn)
        self.gridLayout.addLayout(self.horizontalLayout, 0, 1, 1, 1)
        self.papersize_lbl = QLabel(self)
        self.papersize_lbl.setText(_("Paper size: "))
        self.papersize_lbl.setAlignment(Qt.AlignRight|Qt.AlignTrailing|
                                        Qt.AlignVCenter)
        self.gridLayout.addWidget(self.papersize_lbl, 1, 0, 1, 1)
        self.papersize_btn = QPushButton(self)
        self.papersize_btn.setToolTip(_("Select paper size"))
        self.papersize_btn.setText("210 x 297")
        self.gridLayout.addWidget(self.papersize_btn, 1, 1, 1, 1)
        self.orient_lbl = QLabel(self)
        self.orient_lbl.setText(_("Orientation: "))
        self.orient_lbl.setAlignment(Qt.AlignRight|Qt.AlignTrailing|
                                     Qt.AlignVCenter)
        self.gridLayout.addWidget(self.orient_lbl, 2, 0, 1, 1)
        self.orient_cmb = QComboBox(self)
        self.orient_cmb.addItems([_("Portrait"), _("Landscape")])
        self.gridLayout.addWidget(self.orient_cmb, 2, 1, 1, 1)
        self.color_lbl = QLabel(self)
        self.color_lbl.setText(_("Color: "))
        self.color_lbl.setAlignment(Qt.AlignRight|Qt.AlignTrailing|
                                    Qt.AlignVCenter)
        self.gridLayout.addWidget(self.color_lbl, 3, 0, 1, 1)
        self.color_cmb = QComboBox(self)
        self.color_cmb.addItems([_("Colors"), _("Grayscale")])
        self.gridLayout.addWidget(self.color_cmb, 3, 1, 1, 1)
        self.reso_lbl = QLabel(self)
        self.reso_lbl.setText(_("Resolution: "))
        self.reso_lbl.setAlignment(Qt.AlignRight|Qt.AlignTrailing|
                                   Qt.AlignVCenter)
        self.gridLayout.addWidget(self.reso_lbl, 4, 0, 1, 1)
        self.horizontalLayout_1 = QHBoxLayout()
        self.reso_spb = QSpinBox(self)
        self.reso_spb.setMaximum(9999)
        self.reso_spb.setProperty("value", 100)
        self.reso_spb.setDisplayIntegerBase(10)
        self.reso_spb.setSuffix(" dpi")
        self.horizontalLayout_1.addWidget(self.reso_spb)
        self.apply_reso_btn = QToolButton(self)
        icon3 = QIcon()
        icon3.addPixmap(QPixmap("medias/apply.png"), QIcon.Normal, 
                       QIcon.Off)
        self.apply_reso_btn.setIcon(icon3)
        self.horizontalLayout_1.addWidget(self.apply_reso_btn)
        self.gridLayout.addLayout(self.horizontalLayout_1, 4, 1, 1, 1)
        self.fname_lbl = QLabel(self)
        self.fname_lbl.setText(_("File name: "))
        self.fname_lbl.setAlignment(Qt.AlignRight|Qt.AlignTrailing|
                                    Qt.AlignVCenter)
        self.gridLayout.addWidget(self.fname_lbl, 5, 0, 1, 1)
        self.horizontalLayout_2 = QHBoxLayout()
        self.fname_led = QLineEdit(self)
        self.fname_led.setEnabled(False)
        self.horizontalLayout_2.addWidget(self.fname_led)
        self.browse_btn = QToolButton(self)
        self.browse_btn.setToolTip("Browse")
        self.browse_btn.setText("...")
        self.browse_btn.setEnabled(False)
        self.horizontalLayout_2.addWidget(self.browse_btn)
        self.gridLayout.addLayout(self.horizontalLayout_2, 5, 1, 1, 1)
        self.copies_lbl = QLabel(self)
        self.copies_lbl.setText(_("Copies: "))
        self.copies_lbl.setAlignment(Qt.AlignRight|Qt.AlignTrailing|
                                     Qt.AlignVCenter)
        self.gridLayout.addWidget(self.copies_lbl, 6, 0, 1, 1)
        self.copies_spb = QSpinBox(self)
        self.copies_spb.setMaximum(9999)
        self.copies_spb.setProperty("value", 1)
        self.gridLayout.addWidget(self.copies_spb, 6, 1, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.horizontalLayout_3 = QHBoxLayout()
        sp = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(sp)
        self.abort_btn = QPushButton(self)
        self.abort_btn.setText(_("Cancel"))
        icon1 = QIcon()
        icon1.addPixmap(QPixmap("medias/close.png"), QIcon.Normal, QIcon.Off)
        self.abort_btn.setIcon(icon1)
        self.horizontalLayout_3.addWidget(self.abort_btn)
        self.print_btn = QPushButton(self)
        self.print_btn.setText(_("Print"))
        icon2 = QIcon()
        icon2.addPixmap(QPixmap("medias/apply.png"), QIcon.Normal, QIcon.Off)
        self.print_btn.setIcon(icon2)
        self.horizontalLayout_3.addWidget(self.print_btn)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.grid.addLayout(self.verticalLayout, 0, 0, 1, 1)
        self.set_style()

    def set_style(self):
        self.output_cmb.setStyleSheet(styles.COMBO)
        self.settings_btn.setStyleSheet(styles.TBUTTON)
        self.papersize_btn.setStyleSheet(styles.PUSHBUTTON)
        self.orient_cmb.setStyleSheet(styles.COMBO)
        self.color_cmb.setStyleSheet(styles.COMBO)
        self.reso_spb.setStyleSheet(styles.SPINBOX)
        self.apply_reso_btn.setStyleSheet(styles.TBUTTON)
        self.fname_led.setStyleSheet(styles.LINEEDIT)
        self.browse_btn.setStyleSheet(styles.TBUTTON)
        self.copies_spb.setStyleSheet(styles.SPINBOX)
        self.abort_btn.setStyleSheet(styles.PUSHBUTTON)
        self.print_btn.setStyleSheet(styles.PUSHBUTTON)





