# -*- coding: utf-8 -*-

# printing.py
# this file is part of Oqapy serie 3
#
# Author: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
# Home: http://www.oqapy.eu
# Copyright: 2010-2019 Vincent Vande Vyvre
# Licence: GPL3
#
# printing main class


import sys
import os
import shutil
import logging
lgg = logging.getLogger(__name__)

from PyQt5.QtCore import (QCoreApplication, QMarginsF, QSize, QSizeF, QRectF, 
                          QPointF, QRect, Qt)
from PyQt5.QtGui import QPageSize, QPageLayout, QPageSize, QPainter, QImage, QColor
from PyQt5.QtWidgets import QApplication
from PyQt5.QtPrintSupport import QPrinter, QPrinterInfo

from printing.ui_printing import UiPrinting
from tools.rectsize import RectSize

UNITS = ["mm", "Point", "in", "Pica", "Didot", "Cicero", "Device pixel"]

class Printing:
    """Define the main class of the printing feature.

    """
    def __init__(self, core, images, index):
        """Instanciate the class Printing.

        Args:
        core -- The instance of Core
        images -- The list of the current images in Oqapy
        index -- The index of the current picture
        """
        lgg.info("Call printing dialog")
        self.core = core
        self.images = images
        self.current_folder = ""
        if images:
            self.current_folder = os.path.dirname(images[0])

        self.index = index
        self.no_printer = True
        self.ui = UiPrinting(self)
        self.ui.show()
        self.infos = QPrinterInfo()
        self.printers = self.infos.availablePrinterNames()
        if self.printers:
            self.no_printer = False
            default = self.infos.defaultPrinter()
            if default.isNull():
                self.current_printer_infos = self.printers[0]

            else:
                self.current_printer_infos = default

            name = self.current_printer_infos.printerName()
            self.current_printer = Printer(self.current_printer_infos, self)
            self.ui.printertool.set_printer_parameters(self.current_printer)

        self.ui.previewtool.configure(self.images, self.index)
        self.ui.scene.create_frame(self.current_printer)
        self.render = Render(self)
        self.ui.scene.itemSelected.connect(self.on_selection_changed)

    def printer_parameters_changed(self):
        """Update the configuration of the current printer.

        Called when QPageSetupDialog was executed.
        """
        lgg.info("Update current printer settings")
        self.current_printer.update()
        self.ui.printertool.set_printer_parameters(self.current_printer)

    def change_paper_settings(self):
        """Update the paper frame when paper size changed.

        """
        lgg.info("Update paper size")
        self.ui.viewer.scene().update_frame()

    def change_resolution(self, resolution):
        """Update the resolution of the current printer.

        """
        lgg.info("Update resolution to %s" % resolution)
        self.current_printer.setResolution(resolution)

    def create_new_print(self):
        """Restart with blank sheet.

        """
        lgg.info("Create new blank sheet")
        # FIXME check for previous work
        self.render.set_new_render()
        self.configure_layout_toolbox()

    def add_new_image_item(self, qpixmap=None, fname=None):
        """Add a new image item to the render.

        Args:
        qpixmap -- The instance of the QPixmap
        fname -- The full path of the image file
        """
        name = ""
        if fname is not None:
            name = os.path.basename(fname)

        lgg.info("Add item %s" % name)
        self.render.add_image_item(qpixmap, fname)

    def configure_layout_toolbox(self):
        """Update the layout tool box.

        """
        lgg.info("Reset layouttool")
        self.ui.layouttool.reset()

    def send_to_printer(self, color, copies, path):
        """Send the page render to the printer.

        Args:
        colors -- If 0 print in color else print in grayscale
        copies -- Number of copies
        path -- The path of the pdf file
        """
        lgg.info("Send the job to the printer")
        self.set_buzzy_widget(True)
        content = self.render.build_render(self.ui.scene.page_size)
        self.ui.scene.prepare_to_print()
        if path:
            self.current_printer.print_pdf(color, content, path)

        else:
            self.current_printer.print_page(color, copies, content)

    def on_printing_finished(self):
        """Rebuild the page frame removed for the printing.

        """
        lgg.info("Printing finished")
        self.ui.scene.create_frame(self.current_printer)
        self.set_buzzy_widget(False)

    def set_buzzy_widget(self, state):
        lgg.info("Set buzzy cursor: %s" % state)
        if state:
            QApplication.setOverrideCursor(Qt.WaitCursor)

        else:
            QApplication.restoreOverrideCursor()
        QCoreApplication.processEvents()

    def on_selection_changed(self):
        """Called by the QGraphicsScene.

        """
        lgg.info("Selection changed")
        item = self.ui.scene.selection
        if not item:
            self.configure_layout_toolbox()
            return

        pos = item.pos()
        p = item.pixmap()
        self.ui.layouttool.update_with_item(item, self.ui.scene.pixels_to_mm, 
                                            self.current_printer.margins)

    def _on_orientation_changed(self, orientation):
        """Called by the orientation combobox of the printer toolbox.

        Args:
        orientation -- Paper orientations: 0 = portrait, 1 = landscape
        """
        lgg.info("Orientation changed %s" % orientation)
        self.current_printer.setPageOrientation(orientation)
        self.current_printer.update()


class Printer(QPrinter):
    """Define a printer device.

    """
    def __init__(self, infos, main):
        """Instanciate the printer.

        Args:
        infos -- Instance of QPrinterInfo
        main -- Instance of Printing
        """
        super().__init__(infos, QPrinter.HighResolution)
        self.main = main
        self.model = infos.makeAndModel()
        self.name = infos.printerName()
        self.unit = 0
        self.orientation = 0
        max_ = infos.maximumPhysicalPageSize().size(QPageSize.Millimeter)
        min_ = infos.minimumPhysicalPageSize().size(QPageSize.Millimeter)
        self.max_width = max_.width()
        self.max_height = max_.height()
        self.min_width = min_.width()
        self.min_height = min_.height()
        self.default_size = infos.defaultPageSize().size(QPageSize.Millimeter)
        paper = self.paperRect(QPrinter.Millimeter)
        page = self.pageRect(QPrinter.Millimeter)
        # Margins [left, top, right, bottom]
        self.margins = [page.x(), page.y(), 
                        paper.width() - (page.width() + page.x()),
                        paper.height() - (page.height() + page.y())]

    def update(self):
        lgg.info("Printer update")
        self.orientation = self.pageLayout().orientation()
        self.main.change_paper_settings()

    def print_infos(self):
        # Debugging usage
        inf = QPrinterInfo().printerInfo(self.name)
        print("Name:", inf.printerName())
        print("Description:", inf.description())
        print("Location:", inf.location())
        print("Make:", inf.makeAndModel())
        print("Size max:", inf.maximumPhysicalPageSize().size(QPageSize.Millimeter))
        print("Size min:", inf.minimumPhysicalPageSize().size(QPageSize.Millimeter))
        print("Resolutions:", inf.supportedResolutions())
        print("Custom sizes:", inf.supportsCustomPageSizes())
        print("Resolution:", self.resolution())
        print("Orientation:", self.pageLayout().orientation())
        print("Paper in mm:", self.paperRect(QPrinter.Millimeter))
        print("Paper in pixels:", self.paperRect(QPrinter.DevicePixel))
        print("Page in pixels:", self.pageRect(QPrinter.DevicePixel))

    @property
    def orientation_(self):
        """Return the paper orientation.

        """
        return self.pageLayout().orientation()

    @property
    def paper_size(self):
        """Return the paper size.

        """
        size = self.paperRect(QPrinter.Millimeter)
        return size.width(), size.height()

    @property
    def printable_area(self):
        """Return the page size.

        """
        if not self.unit:
            page = self.pageRect(QPrinter.Millimeter)

        else:
            page = self.pageRect(QPrinter.Inch)

        return page.width(), page.height()

    def set_resolution(self, value):
        """Set the printer resolution.

        Args:
        value -- The value of the resolution
        """
        lgg.info("Printer set resolution %s" % value)
        self.setResolution(value)

    def print_page(self, gray, copies, img):
        """Run a printing job.

        Args:
        gray -- If 1 print in grayscale else print in color
        img -- The content of the print created by the Render
        """
        lgg.info("Printer print with current printer")
        if gray:
            self.setColorMode(QPrinter.GrayScale)

        target = self.paperRect(QPrinter.DevicePixel)
        painter = QPainter()
        painter.begin(self)
        for i in range(copies):
            painter.drawImage(target, img, target)
            # Avoid unusued last blank sheet
            if not i == copies - 1:
                self.newPage()

        painter.end()
        del img
        self.setColorMode(QPrinter.Color)
        self.main.on_printing_finished()

    def print_pdf(self, gray, img, fname):
        """Record the scene into a pdf or postscript format.

        Args:
        gray -- If 1 print in grayscale else print in color
        img -- The content of the print created by the Render
        fname -- The path of the file
        """
        lgg.info("Printer export to PDF\n\t\t%s" % fname)
        self.setOutputFormat(QPrinter.PdfFormat) 
        self.setOutputFileName(fname)
        if gray:
            self.setColorMode(QPrinter.GrayScale)

        target = self.paperRect(QPrinter.DevicePixel)
        painter = QPainter()
        painter.begin(self)
        painter.drawImage(target, img, target)
        painter.end()
        del img
        self.setOutputFormat(QPrinter.NativeFormat)
        self.setOutputFileName('')
        self.setColorMode(QPrinter.Color)
        self.main.on_printing_finished()


class Render:
    """Define the printing render.

    """
    def __init__(self, main):
        """Instanciate the render.

        Args:
        main -- Printing instance
        """
        self.main = main
        self.printer = main.current_printer
        self.view = main.ui.viewer
        self.scene = main.ui.viewer.scene()
        self.items = []
        self.selection = None
        self.selections = []

    def set_new_render(self):
        """Set a new blank sheet.

        """
        lgg.info("Render clean")
        self.scene.remove_all()
        self.items = []
        self.selection = None

    def add_image_item(self, pixmap, fname):
        """Add a new image into the page.

        Args:
        pixmap -- QPixmap instance
        fname -- Full path of the image file
        """
        lgg.info("Render add image")
        self.selection = None
        self.create_image_item(pixmap, fname)

    def create_image_item(self, pixmap, fname):
        """Create an image item.

        Args:
        pixmap -- QPixmap instance
        fname -- Full path of the image file
        """
        lgg.info("Render create image item")
        reso = self.printer.resolution()
        margins = self.printer.margins
        size_max = self.printer.printable_area
        width, height = self.compute_size(pixmap.size(), self.scene.inmargins)
        self.scene.clearSelection()
        pix_item = self.scene.insert_image(pixmap, fname, width, height)
        pix_item.set_level(len(self.items))
        self.items.append(pix_item)

    def remove_item(self, item):
        """Remove an image item.

        Args:
        item -- The QGraphicsItem to be removed
        """
        lgg.info("Render remove item")
        self.items.remove(item)

    def compute_size(self, img, max_):
        """Return the maximum size for an image item.

        Args:
        img -- The size of the source image
        max_ -- The printable area size

        Returns:
        The image size if it is smaller than max_ otherwise max_
        """
        orig = RectSize(img)
        dest = RectSize(max_)
        if orig.is_include(dest):
            return img.width(), img.height()

        return max_

    def build_render(self, size):
        """Paint the page for the printer.

        Args:
        size -- The size of the QGraphicsScene
        """
        lgg.info("Render build the render of %s items" % len(self.items))
        r = self.printer.pageRect(QPrinter.DevicePixel)
        w = int(r.width())
        h = int(r.height())
        s = QSize(w, h)
        background = QImage(s, QImage.Format_ARGB32_Premultiplied)
        background.fill(QColor(0, 0, 0, 0))
        factor = w / size.width
        if size.is_vertical:
            factor = h / size.height

        painter = QPainter()
        painter.begin(background)
        items = {}
        for i in self.items:
            items[i.level] = i

        for k in sorted(items.keys()):
            i = items[k]
            ix = i.pos().x() * factor
            iy = i.pos().y() * factor
            iwidth = i.pixmap_width * factor
            iheight = i.pixmap_height * factor
            target = QRect(int(ix), int(iy), int(iwidth), int(iheight))
            img = QImage(i.path)
            source = QRect(0, 0, img.width(), img.height())
            painter.drawImage(target, img, source)

        del img
        painter.end()
        return background

    def set_item_z_order(self, value, item=None):
        """Change the z-order of an item.

        Args:
        value -- The direction of z: -1 to the background or 1 to the foreground
        item -- The image item
        """
        if item is None:
            if not self.scene.selection:
                return

            item = self.scene.selection

        old = item.level
        new = old + value
        max_ = len(self.items) - 1
        if new < 0 or new > max_:
            return

        for i in self.items:
            if i.level == new:
                # item permutation
                i.set_level(old)
                break

        item.set_level(new)
        self.main.ui.layouttool.configure_order_buttons(new, max_)

    def set_item_foreground(self, item):
        """Move the item to the foreground.

        Args:
        item -- The image item
        """
        max_ = len(self.items) - 1
        while item.level < max_:
            self.set_item_z_order(1, item)

    def set_item_background(self, item):
        """Move the item to the background.

        Args:
        item -- The image item
        """
        while item.level > 0:
            self.set_item_z_order(-1, item)

