# -*- coding: utf-8 -*-

# ui_printing.py
# this file is part of Oqapy serie 3
#
# Author: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
# Home: http://www.oqapy.eu
# Copyright: 2010-2019 Vincent Vande Vyvre
# Licence: GPL3
#
# printing dialog box.

import sys
import os
import logging
lgg = logging.getLogger(__name__)

from PyQt5.QtCore import (Qt, pyqtSignal, QMetaObject, QSize, QRectF, QPointF,
                          QTimer)
from PyQt5.QtGui import (QIcon, QPixmap, QPen, QBrush, QColor, QImage, QCursor,
                         QImageReader, QPainter, QPen, QFont, QTransform)
from PyQt5.QtWidgets import (QDialog, QGridLayout, QVBoxLayout, QHBoxLayout,
                             QSplitter, QSpacerItem, QSizePolicy, QGraphicsView,
                             QLabel, QSpinBox, QPushButton, QGraphicsScene,
                             QWidget, QGraphicsPixmapItem, QGraphicsItem,
                             QGraphicsRectItem, QStyleOptionGraphicsItem,
                             QStyleOptionFocusRect, QApplication, QMenu, 
                             QAction)

import stylesheets as styles
from VWidgets_PyQt5.vtoolpanel import VToolPanel
from printing.previewertoolbox import PreviewerToolBox
from printing.printertoolbox import PrinterToolBox
from printing.layouttoolbox import LayoutToolBox
from printing.alignmenttoolbox import AlignmentToolBox

from tools.rectsize import RectSize


class UiPrinting(QDialog):
    def __init__(self, main, parent=None):
        super().__init__(parent)
        self.resize(1000, 700)
        self.gridLayout = QGridLayout(self)
        self.gridLayout.setContentsMargins(2, 0, 2, 2)
        self.splitter = QSplitter(self)
        self.splitter.setOrientation(Qt.Horizontal)
        self.viewer = GraphicsView(self, self.splitter)
        self.tools = VToolPanel(self.splitter)
        self.tools.setMaximumSize(QSize(400, 16777215))
        self.tools.setTitle("Settings")
        self.tools.setObjectName("tools")
        self.gridLayout.addWidget(self.splitter, 0, 0, 1, 1)
        self.setSizeGripEnabled(True)
        self.printertool = PrinterToolBox(self)
        self.layouttool = LayoutToolBox(self)
        self.alignmenttool = AlignmentToolBox(self)

        self.retranslateUi()
        self.main = main
        self.set_tool_pages()
        self.scene = GraphicsScene(self, self.viewer)
        self.viewer.setScene(self.scene)
        self.new_layout_btn.clicked.connect(self.main.create_new_print)
        self.enable_layout_tools(False)
        self.set_theme()

    def set_tool_pages(self):
        """Sets the tool pages into the tool panel.

        How to add a tool page:
            - Create, if needed, a new tool page with:
                    page = VToolPanel.addToolPage(name, icon)
            - Create a new toolbox to this page with:
                    myNewPage = VToolPanel.addToolBox(page)
            - Set the properties of the box (title, icon, collapsible, ...)
            - Instanciate the tool widget with the page as parent:
                    myWidget = MyWidget(myNewPage)
            - Place the tool widget into the page layout with:
                    myNewPage.layout.addWidget(myWidget)
        """
        self.grp_0 = self.tools.addToolPage('grp_0')
        self.preview_pg = self.tools.addToolBox(0)
        self.preview_pg.setTitle(_('Image preview'))
        self.previewtool = PreviewerToolBox(self, self.preview_pg)
        self.preview_pg.layout.addWidget(self.previewtool.widget)

        self.printer_pg = self.tools.addToolBox(0)
        self.printer_pg.setTitle(_('Printer'))
        self.printer_box = self.printertool.widget
        self.printer_box.setParent(self.printer_pg)
        self.printer_pg.layout.addWidget(self.printer_box)

        self.layout_pg = self.tools.addToolBox(0)
        self.layout_pg.setTitle(_('Page layout'))
        self.new_layout_btn = self.layout_pg.header.add_button(
                                        "medias/blank-sheet.svg",
                                        "new_layout_btn")
        self.new_layout_btn.setToolTip(_("New blank sheet"))
        self.layout_box = self.layouttool.widget
        self.layout_box.setParent(self.layout_pg)
        self.layout_pg.layout.addWidget(self.layout_box)

        self.alignment_pg = self.tools.addToolBox(0)
        self.alignment_pg.setTitle(_("Alignment - Distribution"))
        self.alignment_box = self.alignmenttool.widget
        self.alignment_box.setParent(self.alignment_pg)
        self.alignment_pg.layout.addWidget(self.alignment_box)

        self.tools.setCurrentPage(0)

    def get_viewer_size(self):
        """Return the QGraphicsView size.

        """
        return self.viewer.size_

    def fit_item(self, size):
        """Resize the current item to fit it into the printable area.

        Args:
        size -- int  1: fit to maximum
                     2: fit to page width
                     3: fit to page height
        """
        if size == 1:
            self.scene.fit_item_in_page()

        elif size == 2:
            self.scene.fit_item_to_width()

        elif size == 3:
            self.scene.fit_item_to_height()

    def enable_layout_tools(self, bool_):
        """Enable or disable the layout and alignment tools.

        Args:
        bool_ -- boolean
        """
        lgg.info("Enable layout tools %s" % bool_)
        self.layouttool.active_widget(bool_)
        self.alignmenttool.active_widget(bool_)

    def enable_alignment_tool(self, bool_):
        """Enable or disable the alignment tool.

        Args:
        bool_ -- boolean
        """
        lgg.info("Enable alignment tools %s" % bool_)
        self.alignmenttool.active_widget(bool_)

    def retranslateUi(self):
        self.setWindowTitle(_("Printing - Oqapy"))

    def set_theme(self):
        self.setStyleSheet("""QWidget {background: #181E26;
                        color: #FFFFFF;}""")
        self.viewer.setStyleSheet(styles.VIEWER)
        self.tools.scrollArea.setStyleSheet(styles.SCROLLAREA)
        self.tools.header.setStyleSheet(styles.HEADER)
        self.tools.header.label.setStyleSheet(styles.TITLE)
        self.tools.scrollArea.horizontalScrollBar().setStyleSheet(
                                                    styles.HSCROLLBAR)
        self.tools.scrollArea.verticalScrollBar().setStyleSheet(
                                                    styles.VSCROLLBAR)
        boxes = self.tools.get_boxes()
        for tb in boxes:
            if tb == self.splitter:
                boxes.extend(tb.boxes)
                continue
            tb.header.setStyleSheet(styles.HEADER)
            tb.header.label.setStyleSheet(styles.TITLE)
            try:
                tb.header.collapse_btn.setStyleSheet(styles.EXPAND)
                tb.header.set_expanding_buttons(
                                        'medias/collapse_white_down.png',
                                        'medias/collapse_white_left.png',
                                        'medias/collapse_white_right.png')
            except:
                # Not all tool boxes have a collapse button
                pass
            tb.header.layout.itemAt(0).widget().setStyleSheet(styles.TITLE)



class GraphicsView(QGraphicsView):
    def __init__(self, ui, parent=None):
        super().__init__(parent)
        self.ui = ui
        self.on_resize = -1

    @property
    def size_(self):
        return self.width(), self.height()

    def resizeEvent(self, event):
        if self.on_resize == -1:
            return

        self.on_resize += 1
        timer = QTimer.singleShot(200, self.resizing)
        QGraphicsView.resizeEvent(self, event)

    def resizing(self):
        if self.on_resize:
            self.on_resize -= 1

            if not self.on_resize:
                lgg.info("View on resizing")
                self.scene()._on_view_resized()

    def keyPressEvent(self, event):
        QGraphicsView.keyPressEvent(self, event)

class GraphicsScene(QGraphicsScene):
    itemSelected = pyqtSignal()
    selectionAdded = pyqtSignal(QGraphicsItem)
    def __init__(self, ui, view):
        super().__init__()
        self.ui = ui
        self.view = view
        # Size in unit of printed image
        self.render_sizes = (0, 0)
        self._sizes = (0, 0)
        self.zoom_factor = 1.0
        self.trans_color = QColor(0, 0, 0, 0)
        self.sheet_color = QColor(255, 255, 255, 255)
        self.frame_color = QColor(0, 0, 0, 255)
        self.shadow_color = QColor(0, 0, 0, 100)
        self.selection = None
        self.selected = []
        self.cursors = {}
        self.active_grip = False
        self.grip_pos = None
        self.button = None
        self.current_qimage = None
        self.page_frame = None
        self.selection_frame = None
        self.selection_frame_rescaled = False
        self.on_alignment = False
        self.item_moved = False
        self.selectionChanged.connect(self.on_selection_changed)
        for f in [("cross", "medias/cross_cursor_32.png"),
                  ("right", "medias/double_arrow_90.png"),
                  ("down" , "medias/double_arrow_0.png")]:
            shape = QPixmap(f[1])
            self.cursors[f[0]] = QCursor(shape, -1, -1)

    def create_frame(self, printer):
        self.printer = printer
        view_size = RectSize(self.view.size_)
        paper_size = RectSize(printer.paper_size)
        factor = paper_size.height / paper_size.width
        if paper_size.is_vertical:
            height = view_size.height - 40
            width = height / factor

        else:
            width = view_size.width - 40
            height = width * factor

        if height > view_size.height - 40:
            height = view_size.height - 40
            width = height / factor

        if width > view_size.width - 40:
            width = view_size.width - 40
            height = width * factor

        lgg.info("Scene create frame %s x %s" %(width, height))
        self.page_size = RectSize(width, height)
        mm = max(paper_size.height, paper_size.width)
        pixels = max(height, width)
        self.pixels_to_mm = pixels / mm
        self.draw_frame(self.page_size)
        self.show_page_frame(self.page_size)
        self.view.on_resize = 0

    def draw_frame(self, size):
        back = self.get_background(size)
        painter = QPainter()
        painter.begin(back)
        painter.setPen(QPen(self.frame_color, 1, Qt.SolidLine, 
                            Qt.RoundCap, Qt.MiterJoin))
        painter.drawLine(1.0, size.height-2, 1.0, 1.0)
        painter.drawLine(1.0, 1.0, size.width-2, 1.0)
        painter.setPen(QPen(self.shadow_color, 3, Qt.SolidLine, 
                            Qt.RoundCap, Qt.MiterJoin))
        painter.drawLine(size.width-2, 3.0, size.width-2, size.height)
        painter.drawLine(size.width-2, size.height-2, 3.0, size.height-2)
        margins = self.printer.margins
        # scene margins in pixels
        self.pixelsmargins = [int(m*self.pixels_to_mm) for m in margins]
        if any(margins):
            x = self.pixelsmargins[0]
            y = self.pixelsmargins[1]
            w = size.width - 2 - x - self.pixelsmargins[2]
            h = size.height - 2 - y  - self.pixelsmargins[3]
            painter.setPen(QPen(self.frame_color, 1, Qt.DashDotLine, 
                            Qt.RoundCap, Qt.MiterJoin))
            painter.drawRect(x, y, w, h)

        painter.end()
        # inmargins is the width & height of the printable area
        self.inmargins = (w, h)
        self.page_frame = back

    def get_background(self, size):
        size.resize(2, 2)
        s = size.as_qsize
        background = QImage(s, QImage.Format_ARGB32_Premultiplied)
        background.fill(self.sheet_color)
        return background

    def show_page_frame(self, size):
        self.setSceneRect(0.0, 0.0, size.width, size.height)
        pix = QPixmap(self.page_frame)
        self.paper_frame = self.addPixmap(pix)
        self.paper_frame.setZValue(-1)

    def update_frame(self):
        self.removeItem(self.paper_frame)
        self.create_frame(self.ui.main.current_printer)

    def rotate(self, orig, angle):
        matrix = QTransform()
        pixmap = orig.transformed(matrix.rotate(angle), 
                             Qt.SmoothTransformation)
        return pixmap

    def resize(self, orig, width, size):
        if width:
            return orig.scaledToWidth(size, Qt.SmoothTransformation)

        return orig.scaledToHeight(size, Qt.SmoothTransformation)

    def remove_item(self, item):
        lgg.info("Scene remove item")
        if item == self.selection:
            self.unselect_all()

        self.ui.main.render.remove_item(item)
        self.removeItem(item)

    def remove_all(self):
        lgg.info("Scene remove all items")
        for item in self.items():
            if item == self.paper_frame:
                continue
            self.removeItem(item)

        self.selection = False
        self.selected = []
        self.ui.enable_layout_tools(False)

    def get_size(self):
        """Return size of graphics scene

        """
        return self.sceneRect()

    def get_ratio(self):
        """Return value dpi/mm for the printer resolution

        """
        return self.ui.main.page.resolution / 25.4

    def get_zoom(self):
        """Return current zoom for the page

        """
        return self.sceneRect().width() / self.ui.main.page.size[0]

    def insert_image(self, qimage, path, width, height):
        lgg.info("Scene insert image item")
        self.unselect_all()
        pixmap = ImageItem(qimage, path, width, height)
        self.addItem(pixmap)
        pw = pixmap.pixmap().size()
        x = (self.width() - pw.width()) / 2
        y = (self.height() - pw.height()) / 2
        pixmap.setPos(int(x), int(y))
        self.item_moved = False
        self.on_image_item_selected(pixmap)
        return pixmap

    def on_image_item_selected(self, item):
        lgg.info("Scene item selected %s" % item)
        if not isinstance(item, ImageItem):
            return

        if self.selection is not None:
            if self.selection == item:
                return

            if self.selection_frame == item:
                return

            if isinstance(item, Grip):
                self.active_grip = item
                self.grip_pos = item.pos()
                return

            self.unselect_item()

        item.set_selected()
        self.selection = item
        self.set_selection_frame(item)
        self.ui.layouttool.active_widget(True)
        self.itemSelected.emit()

    def add_item_selected(self, item):
        lgg.info("Scene add item %s to selection" % item)
        if item == self.selection:
            return

        if item in self.selected:
            self.selected.remove(item)
            item.setSelected(False)
            if not self.selected:
                self.ui.enable_alignment_tool(False)

        else:
            self.selected.append(item)
            item.setSelected(True)
            self.selectionAdded.emit(item)
            self.ui.enable_alignment_tool(True)

    def unselect_item(self):
        if not self.selection or self.selection_frame is None:
            return

        lgg.info("Scene unselect")
        self.removeItem(self.selection_frame)
        for g in self.selection.grips:
            self.removeItem(g)

        self.selection.setSelected(False)
        self.selection.unselect()
        self.selection = False
        self.selection_frame = None
        self.ui.layouttool.active_widget(False)

    def unselect_all(self):
        lgg.info("Scene unselect all")
        self.unselect_item()
        for s in self.selected:
            s.setSelected(False)

        self.selected = []

    def set_selection_frame(self, item):
        pos = item.pos()
        size = item.pixmap().size()
        frame = self.build_selection_frame(size)
        self.selection_frame = self.addPixmap(frame)
        self.selection_frame.setParentItem(item)
        self.selection_frame.setPos(-1, -1)
        item.grips = []
        right = (size.width()-16, (size.height()/2)-8)
        r_grip = Grip(0, right, "medias/right_grip.png", self, 
                        self.cursors["right"], item)
        item.grips.append(r_grip)
        down = ((size.width()/2)-8, size.height()-16)
        d_grip = Grip(1, down, "medias/down_grip.png", self, 
                        self.cursors["down"], item)
        item.grips.append(d_grip)

    def build_selection_frame(self, size, width=None, height=None):
        #lgg.info("Scene build selection frame")
        if size:
            width, height = size.width(), size.height()

        frame = QPixmap(width+2, height+2)
        rect = QRectF(0.0, 0.0, width+2, height+2)
        frame.fill(self.trans_color)
        painter = QPainter()
        painter.begin(frame)
        painter.setPen(QPen(QColor("#FF0000"), 4, Qt.DashDotLine, 
                            Qt.RoundCap, Qt.MiterJoin))
        painter.drawRect(rect)
        painter.end()
        return frame

    def change_grip(self, grip):
        lgg.info("Scene set active grip %s" % grip)
        self.active_grip = grip

    def rescale_selection_frame(self, pos):
        #lgg.info("Scene rescale selection frame")
        fsize = self.selection_frame.pixmap().size()
        width, height = fsize.width() - 2, fsize.height() - 2
        offset = self.selection.pos()
        if not self.active_grip.idx:
            # Horizontal rescale
            width = pos.x() - offset.x()
            if self.ui.layouttool.is_proportionnal:
                height = width * self.selection.width_ratio

        else:
            # Vertical rescale
            height = pos.y() - offset.y()
            if self.ui.layouttool.is_proportionnal:
                width = height * self.selection.height_ratio

        self.selection_frame.setParentItem(None)
        self.removeItem(self.selection_frame)
        frame = self.build_selection_frame(False, width, height)
        self.selection_frame = self.addPixmap(frame)
        self.selection_frame.setParentItem(self.selection)
        self.selection_frame_rescaled = (self.selection.pos(), width, height)
        return width, height

    def move_grips(self, borders):
        right, bottom = borders
        offset = self.selection.pos()
        left, top = offset.x(), offset.y()
        self.selection.grips[0].setPos(right-15, (bottom/2)-8)
        self.selection.grips[1].setPos((right/2)-8, bottom-15)

    def rescale_item_from_selection(self):
        lgg.info("Scene rescale item from selection")
        if self.selection and self.selection_frame_rescaled:
            qimg = self.get_qimage(self.selection.path)
            if not qimg.isNull():
                # FIXME display error message
                self.selection.prepareGeometryChange()
                # selection_frame_rescaled : (pos, width, height)
                width = self.selection_frame_rescaled[1]
                if self.ui.layouttool.is_proportionnal:
                    pix = QPixmap(qimg.scaledToWidth(width, 
                                  Qt.SmoothTransformation))

                else:
                    height = self.selection_frame_rescaled[2]
                    pix = QPixmap(qimg.scaled(width, height, 
                                  Qt.IgnoreAspectRatio, 
                                  Qt.SmoothTransformation))

                self.selection.setPixmap(pix)
                self.ui.layouttool.update_with_item(self.selection,
                                                    self.pixels_to_mm)
                self.ui.layouttool.set_preset_size_index(0)

    def resize_item(self, width, height=0):
        lgg.info("Scene resize item %s, %s" %(width, height))
        if self.selection:
            item = self.selection
            qimg = self.get_qimage(self.selection.path)
            self.unselect_item()
            width *= self.pixels_to_mm
            if not qimg.isNull():
                if not height:
                    pix = QPixmap(qimg.scaledToWidth(width, 
                                  Qt.SmoothTransformation))

                else:
                    height *= self.pixels_to_mm
                    pix = QPixmap(qimg.scaled(width, height, 
                                  Qt.IgnoreAspectRatio, 
                                  Qt.SmoothTransformation))

                item.prepareGeometryChange()
                item.setPixmap(pix)
                self.on_image_item_selected(item)
                self.ui.layouttool.set_preset_size_index(0)

    def resize_all_items(self, factor):
        lgg.info("Scene resize all")
        self.unselect_all()
        for item in self.ui.main.render.items:
            size = item.size_rect
            qimg = self.get_qimage(item.path)
            if size.is_vertical:
                pix = QPixmap(qimg.scaledToHeight(size.height*factor,
                                  Qt.SmoothTransformation))

            else:
                pix = QPixmap(qimg.scaledToWidth(size.width*factor, 
                              Qt.SmoothTransformation))

            item.prepareGeometryChange()
            item.setPixmap(pix)


    def fit_item_in_page(self):
        lgg.info("Scene fit in page")
        if self.selection:
            width, height = self.inmargins
            size = self.selection.pixmap().size()
            if height * (size.width() / width) < size.height():
                self.fit_item_to_height()

            else:
                self.fit_item_to_width()

    def fit_item_to_width(self):
        lgg.info("Scene fit to width")
        if self.selection:
            item = self.selection
            qimg = self.get_qimage(self.selection.path)
            r = self.ui.layouttool.item_resized
            self.unselect_item()
            if r:
                # unselect_item() call a reset of the tool box
                self.ui.layouttool.item_resized = True

            if not qimg.isNull():
                # FIXME display error message
                item.prepareGeometryChange()
                width, _ = self.inmargins
                if self.ui.layouttool.is_proportionnal:
                    pix = QPixmap(qimg.scaledToWidth(width, 
                                  Qt.SmoothTransformation))

                else:
                    height = item.pixmap().height()
                    pix = QPixmap(qimg.scaled(width, height, 
                                  Qt.IgnoreAspectRatio, 
                                  Qt.SmoothTransformation))

                item.setPixmap(pix)
                self.center_item(item, pix.size())
                self.on_image_item_selected(item)

    def fit_item_to_height(self):
        lgg.info("Scene fit height")
        if self.selection:
            item = self.selection
            qimg = self.get_qimage(self.selection.path)
            r = self.ui.layouttool.item_resized
            self.unselect_item()
            if r:
                # unselect_item() call a reset of the tool box
                self.ui.layouttool.item_resized = True

            if not qimg.isNull():
                # FIXME display error message
                item.prepareGeometryChange()
                _, height = self.inmargins
                if self.ui.layouttool.is_proportionnal:
                    pix = QPixmap(qimg.scaledToHeight(height, 
                                  Qt.SmoothTransformation))

                else:
                    width = item.pixmap().width()
                    pix = QPixmap(qimg.scaled(width, height, 
                                  Qt.IgnoreAspectRatio, 
                                  Qt.SmoothTransformation))

                item.setPixmap(pix)
                self.center_item(item, pix.size())
                self.on_image_item_selected(item)

    def center_item(self, item, size):
        lgg.info("Scene center item")
        scene = self.sceneRect()
        x = (scene.width() - size.width()) / 2
        y = (scene.height() - size.height()) / 2
        item.setPos(x, y)
        self.ui.layouttool.update_with_item(item, self.pixels_to_mm)

    def move_item(self, side):
        """Move the selected item."""
        lgg.info("Scene move item")
        size = self.selection.pixmap().size()
        srect = self.sceneRect()
        if side == 1:
            # Top left
            pos = QPointF(self.pixelsmargins[0], self.pixelsmargins[1])

        elif side == 2:
            # Top right
            pos = QPointF(srect.width()-self.pixelsmargins[2]-size.width(), 
                          self.pixelsmargins[1])

        elif side == 3:
            # Center
            self.center_item(self.selection, size)
            self.on_alignment = True
            return

        elif side == 4:
            # Bottom left
            pos = QPointF(self.pixelsmargins[0], 
                         srect.height()-size.height()-self.pixelsmargins[1])

        elif side == 5:
            # Bottom right
            pos = QPointF(srect.width()-self.pixelsmargins[2]-size.width(),
                            srect.height()-size.height()-self.pixelsmargins[1])

        else:
            return

        self.on_alignment = True
        self.selection.setPos(pos)
        self.ui.layouttool.update_with_item(self.selection, self.pixels_to_mm)

    def slip_items(self, x, y):
        lgg.info("Scene slip item")
        if self.selection:
            pos = self.selection.pos()
            self.selection.setPos(pos.x()+x, pos.y()+y)

        for item in self.selected:
            pos = item.pos()
            item.setPos(pos.x()+x, pos.y()+y)

    def change_item_z_order(self, item, change):
        lgg.info("Scene change z order to %s" % change)
        if change == "up":
            self.ui.main.render.set_item_z_order(1, item)

        elif change == "down":
            self.ui.main.render.set_item_z_order(-1, item)

        elif change == "foreground":
            self.ui.main.render.set_item_foreground(item)

        elif change == "background":
            self.ui.main.render.set_item_background(item)

    def set_item_top_position(self, top):
        pos = self.selection.pos()
        self.selection.setPos(pos.x(), top*self.pixels_to_mm)
        self.ui.layouttool.update_with_item(self.selection, self.pixels_to_mm)

    def set_item_right_position(self, right):
        right *= self.pixels_to_mm
        pos = self.selection.pos()
        swidth = self.sceneRect().width()
        width = self.selection.pixmap().width()
        x = swidth - right - width
        self.selection.setPos(x, pos.y())
        self.ui.layouttool.update_with_item(self.selection, self.pixels_to_mm)

    def set_item_bottom_position(self, bottom):
        bottom *= self.pixels_to_mm
        pos = self.selection.pos()
        height = self.selection.pixmap().height()
        sheight = self.sceneRect().height()
        y = sheight - bottom - height
        self.selection.setPos(pos.x(), y)
        self.ui.layouttool.update_with_item(self.selection, self.pixels_to_mm)

    def set_item_left_position(self, left):
        pos = self.selection.pos()
        self.selection.setPos(left*self.pixels_to_mm, pos.y())
        self.ui.layouttool.update_with_item(self.selection, self.pixels_to_mm)

    def align_selection(self, mode):
        lgg.info("Scene align mode %s" % mode)
        if not self.selected:
            return

        self.on_alignment = True
        if mode in (1, 4):
            self.align_right_sides(mode)

        elif mode in (2, 5):
            self.align_left_sides(mode)

        elif mode == 3:
            self.align_vertical_symetry()

        elif mode in (6, 9):
            self.align_bottom_sides(mode)

        elif mode in (7, 10):
            self.align_top_sides(mode)

        elif mode == 8:
            self.align_horizontal_symetry()

    def align_right_sides(self, mode):
        align = self.selection.pos().x()
        if mode == 4:
            align += self.selection.pixmap().width()

        for select in self.selected:
            pos = select.pos()
            width = select.pixmap().width()
            x = align - width
            select.setPos(x, pos.y())
            
    def align_left_sides(self, mode):
        align = self.selection.pos().x()
        if mode == 5:
            align += self.selection.pixmap().width()

        for select in self.selected:
            pos = select.pos()
            select.setPos(align, pos.y())

    def align_vertical_symetry(self):
        x = self.selection.pos().x()
        center = self.selection.pixmap().width() / 2 + x
        for select in self.selected:
            pos = select.pos()
            x = center - select.pixmap().width() / 2
            select.setPos(x, pos.y())
            
    def align_bottom_sides(self, mode):
        align = self.selection.pos().y()
        if mode == 9:
            align += self.selection.pixmap().height()

        for select in self.selected:
            pos = select.pos()
            y = align - select.pixmap().height()
            select.setPos(pos.x(), y)
        
    def align_top_sides(self, mode):
        align = self.selection.pos().y()
        if mode == 10:
            align += self.selection.pixmap().height()

        for select in self.selected:
            pos = select.pos()
            select.setPos(pos.x(), align)

    def align_horizontal_symetry(self):
        y = self.selection.pos().y()
        center = self.selection.pixmap().height() / 2 + y
        for select in self.selected:
            pos = select.pos()
            y = center - select.pixmap().height() / 2
            select.setPos(pos.x(), y)

    def distribute_selection(self, mode):
        if not self.selected:
            return

        self.on_alignment = True
        funcs = [self.distribute_left_sides,
                 self.distribute_vertical_axis,
                 self.distribute_right_sides,
                 self.distribute_horizontal_gap,
                 self.distribute_bottom_sides,
                 self.distribute_horizontal_axis,
                 self.distribute_top_sides,
                 self.distribute_vertical_gap]
        funcs[mode-11]()

    def distribute_left_sides(self):
        items = self.selected[:]
        items.append(self.selection)
        items.sort(key=lambda item: item.pos().x())
        start = items[0].pos().x()
        end = items[-1].pos().x()
        gap = (end - start) / (len(items) - 1)
        x = start + gap
        for item in items[1:-1]:
            y = item.pos().y()
            item.setPos(x, y)
            x += gap

    def distribute_vertical_axis(self):
        items = self.selected[:]
        items.append(self.selection)
        items.sort(key=lambda item: item.pos().x()+(item.pixmap().width()/2))
        start = items[0].pos().x() + (items[0].pixmap().width()/2)
        end = items[-1].pos().x() + (items[-1].pixmap().width()/2)
        gap = (end - start) / (len(items) - 1)
        x = start + gap
        for item in items[1:-1]:
            y = item.pos().y()
            item.setPos(x-item.pixmap().width()/2, y)
            x += gap

    def distribute_right_sides(self):
        items = self.selected[:]
        items.append(self.selection)
        items.sort(key=lambda item: item.pos().x())
        start = items[0].pos().x() + items[0].pixmap().width()
        end = items[-1].pos().x() + items[-1].pixmap().width()
        gap = (end - start) / (len(items) - 1)
        x = start + gap
        for item in items[1:-1]:
            y = item.pos().y()
            item.setPos(x-item.pixmap().width(), y)
            x += gap

    def distribute_horizontal_gap(self):
        items = self.selected[:]
        items.append(self.selection)
        items.sort(key=lambda item: item.pos().x())
        start = items[0].pos().x() + items[0].pixmap().width()
        inner = items[-1].pos().x() - start
        for item in items[1:-1]:
            inner -= item.pixmap().width()

        gap = inner / (len(items) - 1)
        x = start + gap
        for item in items[1:-1]:
            pos = item.pos()
            item.setPos(x, pos.y())
            x += item.pixmap().width() + gap

    def distribute_bottom_sides(self):
        items = self.selected[:]
        items.append(self.selection)
        items.sort(key=lambda item: item.pos().y())
        start = items[0].pos().y() + items[0].pixmap().height()
        end = items[-1].pos().y() + items[-1].pixmap().height()
        gap = (end - start) / (len(items) - 1)
        y = start + gap
        for item in items[1:-1]:
            pos = item.pos()
            item.setPos(pos.x(), y-item.pixmap().height())
            y += gap

    def distribute_horizontal_axis(self):
        items = self.selected[:]
        items.append(self.selection)
        items.sort(key=lambda item: item.pos().y())
        start = items[0].pos().y() + (items[0].pixmap().height()/2)
        end = items[-1].pos().y() + (items[-1].pixmap().height()/2)
        gap = (end - start) / (len(items) - 1)
        y = start + gap
        for item in items[1:-1]:
            x = item.pos().x()
            item.setPos(x, y-item.pixmap().height()/2)
            y += gap

    def distribute_top_sides(self):
        items = self.selected[:]
        items.append(self.selection)
        items.sort(key=lambda item: item.pos().y())
        start = items[0].pos().y()
        end = items[-1].pos().y()
        gap = (end - start) / (len(items) - 1)
        y = start + gap
        for item in items[1:-1]:
            x = item.pos().x()
            item.setPos(x, y)
            y += gap

    def distribute_vertical_gap(self):
        items = self.selected[:]
        items.append(self.selection)
        items.sort(key=lambda item: item.pos().y())
        start = items[0].pos().y() + items[0].pixmap().height()
        end = items[-1].pos().y()
        inner = end - start
        for item in items[1:-1]:
            inner -= item.pixmap().height()

        gap = inner / (len(items) - 1)
        y = start + gap
        for item in items[1:-1]:
            x = item.pos().x()
            item.setPos(x, y)
            y += item.pixmap().height() + gap

    def get_qimage(self, fname):
        if self.ui.previewtool.current_image == fname:
            return self.ui.previewtool.qimage

        return QImage(fname)

    def on_item_moved(self, item):
        self.item_moved = item
        self.ui.layouttool.update_with_item(item, self.pixels_to_mm, True)

    def on_selection_changed(self):
        items = self.selectedItems()
        if not items:
            return

    def prepare_to_print(self):
        self.unselect_all()
        self.removeItem(self.paper_frame)

    def display_contextual_menu(self, event):
        p = event.scenePos()
        item = self.itemAt(p, QTransform())
        if item:
            max_ = len(self.ui.main.render.items) - 1
            self.menu = ItemContextualMenu(item, max_, event.screenPos())

    def mousePressEvent(self, event):
        """
        Qt::NoModifier	0x00000000	        No modifier
        Qt::ShiftModifier	0x02000000	    Shift key
        Qt::ControlModifier	0x04000000	    Ctrl key
        Qt::AltModifier	0x08000000	        Alt key
        Qt::MetaModifier	0x10000000	    Meta key
        Qt::KeypadModifier	0x20000000	    keypad button
        Qt::GroupSwitchModifier	0x40000000	X11 only. A Mode_switch key on 
                                            the keyboard is pressed.
        """
        p = event.scenePos()
        item = self.itemAt(p, QTransform())
        button = event.button()
        mods = event.modifiers()
        if button == Qt.LeftButton:
            if self.active_grip:
                self.button = button
                QGraphicsScene.mousePressEvent(self, event)
                return

            if item is not None:
                if self.selection == item:
                    QGraphicsScene.mousePressEvent(self, event)

                elif self.selection and mods == Qt.ControlModifier:
                    self.add_item_selected(item)

                else:
                    self.unselect_all()
                    self.on_image_item_selected(item)

                return

            else:
                self.unselect_item()
                self.selected = []
                self.item_moved = False

        QGraphicsScene.mousePressEvent(self, event)

    def mouseReleaseEvent(self, event):
        if self.selection_frame_rescaled:
            self.rescale_item_from_selection()

        elif self.on_alignment:
            # Ignore alignment-distribution movement
            self.on_alignment = False

        elif self.item_moved:
            # The item was unselected
            self.on_image_item_selected(self.item_moved)
            self.item_moved = False
            p = self.selection.pos()
            s = self.selection.pixmap().size()

        self.selection_frame_rescaled = False
        self.button = None
        QGraphicsScene.mouseReleaseEvent(self, event)

    def mouseMoveEvent(self, event):
        if self.button == Qt.LeftButton and self.active_grip:
            size = self.rescale_selection_frame(event.scenePos())
            self.move_grips(size)

        else:
            QGraphicsScene.mouseMoveEvent(self, event)

    def contextMenuEvent(self, event):
        self.display_contextual_menu(event)
        event.accept()

    def keyPressEvent(self, event):
        """
        Qt::Key_Left	 
        Qt::Key_Up	 
        Qt::Key_Right	 
        Qt::Key_Down
        """
        if not self.selection:
            QGraphicsView.keyPressEvent(self, event)

        else:
            mod = QApplication.keyboardModifiers()
            key = event.key()
            offset = 1
            if mod == Qt.ControlModifier:
                offset = 5

            if key == Qt.Key_Left:
                self.slip_items(-offset, 0)

            elif key == Qt.Key_Up:
                self.slip_items(0, -offset)

            elif key == Qt.Key_Right:
                self.slip_items(offset, 0)

            elif key == Qt.Key_Down:
                self.slip_items(0, offset)

    def _on_view_resized(self):
        self.removeItem(self.paper_frame)
        old = RectSize(self.sceneRect().size())
        self.create_frame(self.ui.main.current_printer)
        new =  RectSize(self.sceneRect().size())
        factor = new.width / old.width
        if new.is_vertical:
            factor = new.height / old.height

        self.resize_all_items(factor)


class ImageItem(QGraphicsPixmapItem):
    def __init__(self, qimage, path, width, height, parent=None):
        super().__init__(parent)
        self.path = path
        self.name = os.path.basename(path)
        self.is_selected = False
        self.grips = []
        self.setFlags(QGraphicsItem.ItemIsMovable | 
                      QGraphicsItem.ItemIsSelectable |
                      QGraphicsItem.ItemSendsGeometryChanges)
        self.setTransformationMode(Qt.SmoothTransformation)
        pixmap = QPixmap(qimage.scaled(width, height, Qt.KeepAspectRatio, 
                                Qt.SmoothTransformation))
        self.setPixmap(pixmap)

    @property
    def height_ratio(self):
        # width = height * height_ratio
        return self.pixmap().width() / self.pixmap().height()

    @property
    def width_ratio(self):
        # height = width * width_ratio
        return self.pixmap().height() / self.pixmap().width()

    @property
    def pixmap_width(self):
        return self.pixmap().width()

    @property
    def pixmap_height(self):
        return self.pixmap().height()

    @property
    def size_rect(self):
        return RectSize(self.pixmap_width, self.pixmap_height)

    @property
    def level(self):
        return self.zValue()

    def set_level(self, level):
        self.setZValue(level)

    def move_up(self):
        self.scene().change_item_z_order(self, "up")

    def move_down(self):
        self.scene().change_item_z_order(self, "down")

    def move_foreground(self):
        self.scene().change_item_z_order(self, "foreground")

    def move_background(self):
        self.scene().change_item_z_order(self, "background")

    def delete(self):
        self.scene().remove_item(self)

    def itemChange(self, change, value):
        if change == 0:
            self.scene().on_item_moved(self)
            return value

        else:
            return QGraphicsItem.itemChange(self, change, value)

    def set_selected(self):
        self.is_selected = True

    def unselect(self):
        self.grips = []

class Grip(QGraphicsPixmapItem):
    def __init__(self, idx, pos, icon, scene, cursor=None, parent=None):
        super().__init__(parent)
        self.scene_ = scene
        self.idx = idx
        self._cursor = cursor
        pix = QPixmap(icon)
        self.setPixmap(pix)
        self.setOpacity(0.5)
        self.setPos(pos[0], pos[1])
        self.setFlags(QGraphicsItem.ItemIgnoresParentOpacity)
        self.setAcceptHoverEvents(True)

    def hoverEnterEvent(self, event):
        self.setCursor(self._cursor)
        self.setOpacity(1.0)
        self.scene_.change_grip(self)

    def hoverLeaveEvent(self, event):
        self.scene_.change_grip(False)
        self.setOpacity(0.5)
        event.accept()


class ItemContextualMenu(QMenu):
    def __init__(self, item, number, pos):
        super().__init__(None)
        self.title = QAction(None)
        self.title.setText(item.name)
        self.addAction(self.title)
        self.up = QAction(None)
        self.up.setText(_("Up"))
        self.addAction(self.up)
        self.down = QAction(None)
        self.down.setText(_("Down"))
        self.addAction(self.down)
        self.foreground = QAction(None)
        self.foreground.setText(_("Up to foreground"))
        self.addAction(self.foreground)
        self.background = QAction(None)
        self.background.setText(_("Down to background"))
        self.addAction(self.background)
        self.remove = QAction(None)
        self.remove.setText(_("Delete"))
        self.addAction(self.remove)
        self.up.triggered.connect(item.move_up)
        self.down.triggered.connect(item.move_down)
        self.foreground.triggered.connect(item.move_foreground)
        self.background.triggered.connect(item.move_background)
        self.remove.triggered.connect(item.delete)
        level = item.level
        if level == 0:
            self.down.setEnabled(False)
            self.background.setEnabled(False)

        if not number or level == number:
            self.up.setEnabled(False)
            self.foreground.setEnabled(False)

        self.popup(pos)





